"""A de novo peptide sequencing model."""
import heapq
import logging
import operator
from typing import Any, Dict, List, Optional, Set, Tuple, Union
import depthcharge.masses
import einops
import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils.tensorboard import SummaryWriter
from depthcharge.components import ModelMixin, PeptideDecoder, SpectrumEncoder
from torch.nn.functional import one_hot
import math
import re
from . import evaluate
from ..data import ms_io
from collections import defaultdict
logger = logging.getLogger("diffnovo_dia")

def cosine_beta_schedule(timesteps: int, s: float = 0.008) -> torch.FloatTensor:
#def cosine_beta_schedule(timesteps: int, s: float = 0.0008) -> torch.FloatTensor:
    """Cosine schedule as proposed in https://arxiv.org/abs/2102.09672 .

    Returns alpha parameters, NOT Beta
    """
    steps = timesteps + 1
    x = torch.linspace(0, timesteps, steps)
    alphas_cumprod = torch.cos(((x / timesteps) + s) / (1 + s) * torch.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    alphas = alphas_cumprod[1:] / alphas_cumprod[:-1]
    alphas = torch.clamp(alphas, 0.001, 1.0)
    return torch.sqrt(alphas)


class Spec2Pep(pl.LightningModule, ModelMixin):
    """
    A Transformer model for de novo peptide sequencing.

    Use this model in conjunction with a pytorch-lightning Trainer.

    Parameters
    ----------
    dim_model : int
        The latent dimensionality used by the transformer model.
    n_head : int
        The number of attention heads in each layer. ``dim_model`` must be
        divisible by ``n_head``.
    dim_feedforward : int
        The dimensionality of the fully connected layers in the transformer
        model.
    n_layers : int
        The number of transformer layers.
    dropout : float
        The dropout probability for all layers.
    dim_intensity : Optional[int]
        The number of features to use for encoding peak intensity. The remaining
        (``dim_model - dim_intensity``) are reserved for encoding the m/z value.
        If ``None``, the intensity will be projected up to ``dim_model`` using a
        linear layer, then summed with the m/z encoding for each peak.
    custom_encoder : Optional[Union[SpectrumEncoder, PairedSpectrumEncoder]]
        A pretrained encoder to use. The ``dim_model`` of the encoder must be
        the same as that specified by the ``dim_model`` parameter here.
    max_length : int
        The maximum peptide length to decode.
    residues: Union[Dict[str, float], str]
        The amino acid dictionary and their masses. By default ("canonical) this
        is only the 20 canonical amino acids, with cysteine carbamidomethylated.
        If "massivekb", this dictionary will include the modifications found in
        MassIVE-KB. Additionally, a dictionary can be used to specify a custom
        collection of amino acids and masses.
    max_charge : int
        The maximum precursor charge to consider.
    precursor_mass_tol : float, optional
        The maximum allowable precursor mass tolerance (in ppm) for correct
        predictions.
    isotope_error_range : Tuple[int, int]
        Take into account the error introduced by choosing a non-monoisotopic
        peak for fragmentation by not penalizing predicted precursor m/z's that
        fit the specified isotope error:
        `abs(calc_mz - (precursor_mz - isotope * 1.00335 / precursor_charge))
        < precursor_mass_tol`
    n_beams: int
        Number of beams used during beam search decoding.
    n_log : int
        The number of epochs to wait between logging messages.
    tb_summarywriter: Optional[str]
        Folder path to record performance metrics during training. If ``None``,
        don't use a ``SummaryWriter``.
    warmup_iters: int
        The number of warm up iterations for the learning rate scheduler.
    max_iters: int
        The total number of iterations for the learning rate scheduler.
    out_writer: Optional[str]
        The output writer for the prediction results.
    **kwargs : Dict
        Additional keyword arguments passed to the Adam optimizer.
    """

    def __init__(
        self,
        dim_model: int = 512,
        vocab_size=26,
        win_size=10,
        num_ions=8,
        num_neighbor=5,
        n_head: int = 8,
        dim_feedforward: int = 1024,
        n_layers: int = 9,
        dropout: float = 0.0,
        dim_intensity: Optional[int] = None,
        custom_encoder: Optional[SpectrumEncoder] = None,
        max_length: int = 100,
        residues: Union[Dict[str, float], str] = "canonical",
        max_charge: int = 5,
        precursor_mass_tol: float = 50,
        isotope_error_range: Tuple[int, int] = (0, 1),
        n_beams: int = 5,
        n_log: int = 10,
        time_steps: int = 7,
        tb_summarywriter: Optional[
            torch.utils.tensorboard.SummaryWriter
        ] = None,
        warmup_iters: int = 100_000,
        max_iters: int = 600_000,
        out_writer: Optional[ms_io.MztabWriter] = None,
        **kwargs: Dict,
    ):
        super().__init__()

        # Build the model.
        if custom_encoder is not None:
            self.encoder = custom_encoder
        else:
            self.encoder = SpectrumEncoder(
                dim_model=dim_model,
                vocab_size=vocab_size,
                win_size=win_size,
                num_ions=num_ions,
                num_neighbor=num_neighbor,
                n_head=n_head,
                dim_feedforward=dim_feedforward,
                n_layers=n_layers,
                dropout=dropout,
                dim_intensity=dim_intensity,
            )
        self.decoder = PeptideDecoder(
            dim_model=dim_model,
            n_head=n_head,
            dim_feedforward=dim_feedforward,
            n_layers=n_layers,
            dropout=dropout,
            residues=residues,
            max_charge=max_charge,
        )
        self.softmax = torch.nn.Softmax(2)
        self.celoss = torch.nn.CrossEntropyLoss(ignore_index=0)
        # Optimizer settings.
        self.warmup_iters = warmup_iters
        self.max_iters = max_iters
        self.opt_kwargs = kwargs

        # Data properties.
        self.max_length = max_length
        self.residues = residues
        self.precursor_mass_tol = precursor_mass_tol
        self.isotope_error_range = isotope_error_range
        self.n_beams = n_beams
        self.peptide_mass_calculator = depthcharge.masses.PeptideMass(
            self.residues
        )
        self.stop_token = self.decoder._aa2idx["$"]

        # Logging.
        self.n_log = n_log
        self._history = []
        if tb_summarywriter is not None:
            self.tb_summarywriter = SummaryWriter(tb_summarywriter)
        else:
            self.tb_summarywriter = tb_summarywriter

        # Output writer during predicting.
        self.out_writer = out_writer

        #diffusion model 
        self.time_steps = time_steps
        self.residues = residues
        self.beta_min = 0.01  # Minimum noise level
        self.beta_max = 0.2

        # precision - recall
        self.aa_precisions = []
        self.aa_recalls = []
        self.counter = 0
        self.pep_recalls = []
        self.psm = defaultdict()

        #diffusion model
        self.time_steps = time_steps
        self.residues = residues
        self.beta_min = 0.01  # Minimum noise level
        self.beta_max = 0.2

    # Cosine scheduler for beta values
    def cosine_beta_schedule(self, t, num_steps, beta_min, beta_max):
        cosine_decay = 0.5 * (1 + math.cos(math.pi * t / num_steps))
        return beta_min + (beta_max - beta_min) * (1 - cosine_decay)

    # Transition matrix at each time step
    def get_transition_matrix(self, t):
        vocab_size = self.decoder.vocab_size + 1
        beta_t = self.cosine_beta_schedule(t, self.time_steps, self.beta_min, self.beta_max)
        uniform_matrix = torch.ones((vocab_size, vocab_size)) / vocab_size
        Q_t = (1 - beta_t) * torch.eye(vocab_size) + beta_t * uniform_matrix
        return Q_t

    def forward_diffusion(self, x_0):
        x_t = one_hot(x_0, num_classes=self.decoder.vocab_size + 1).float()
        for t in range(self.time_steps):
            Q_t = self.get_transition_matrix(t).to(x_t.device)
            x_t = torch.matmul(x_t, Q_t)
        return x_t

    def reverse_denoising(self, x_t, spectra, ms1, ms2_b_ion, ms2_y_ion, precursors):
        batch_size = spectra.shape[0]
        for t in reversed(range(self.time_steps)):
            times = (t * torch.ones((batch_size,))).long().to(spectra.device)
            scores, tokens = self.decoder(x_t, times, precursors, *self.encoder(spectra, ms1, ms2_b_ion, ms2_y_ion))

        return scores, tokens

    def _extract_predictions(self, sample: torch.LongTensor) -> list[list[str]]:
        output = []
        for sequence in sample:
            tokens = sequence.tolist()
            if self.residues.eos_index in sequence:
                peptide = tokens[: tokens.index(self.residues.eos_index)]
            else:
                peptide = tokens
            output.append(self.residues.decode(peptide))
        return output

    def forward(self, spectra: torch.Tensor, ms1: torch.Tensor, ms2_b_ion: torch.Tensor, ms2_y_ion: torch.Tensor, precursors: torch.Tensor
    ) -> Tuple[List[List[str]], torch.Tensor]:
        """
        Predict peptide sequences for a batch of MS/MS spectra.

        Parameters
        ----------
        spectra : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra for which to predict peptide sequences.
            Axis 0 represents an MS/MS spectrum, axis 1 contains the peaks in
            the MS/MS spectrum, and axis 2 is essentially a 2-tuple specifying
            the m/z-intensity pair for each peak. These should be zero-padded,
            such that all of the spectra in the batch are the same length.
         ms1: torch.Tensor of shape (n_spectra, n_neighbors)
             the elusion profile for each precursor ion
        ms2_b_ion: torch.Tensor of shape (n_spectra, seq_len, vocab_size,n_ion_type, n_neighbors, window_size)
                  prodcut b ion peptide
        ms2_y_ion: torch.Tensor of shape (n_spectra, seq_len, vocab_size,n_ion_type, n_neighbors, window_size)
                   product y ion peptide
        precursors : torch.Tensor of size (n_spectra, 3)
            The measured precursor mass (axis 0), precursor charge (axis 1), and
            precursor m/z (axis 2) of each MS/MS spectrum.

        Returns
        -------
        peptides : List[List[str]]
            The predicted peptide sequences for each spectrum.
        aa_scores : torch.Tensor of shape (n_spectra, length, n_amino_acids)
            The individual amino acid scores for each prediction.
        """
        aa_beam_scores, beam_tokens = self.beam_search_decode(
            spectra.to(self.encoder.device),
            ms1.to(self.encoder.device),
            ms2_b_ion.to(self.encoder.device),
            ms2_y_ion.to(self.encoder.device),
            precursors.to(self.decoder.device),
        )

        # Reverse process, target is the original sequence batch_x
        diff_scores, diff_tokens = self.reverse_denoising(beam_tokens, spectra, ms1, ms2_b_ion, ms2_y_ion, precursors)
        post_corrected_pep, post_positional_score = [],[]
        for diff_score, mass, charge in zip (diff_scores, precursors[:,2], precursors[:,1]):
           peptide, pos, positional_socre = self.beam_post_correction(diff_score, mass, charge )
           if '$' in peptide: peptide.remove('$')
           post_corrected_pep.append(peptide)
           post_positional_score.append(positional_socre)

        return [self.decoder.detokenize(t) for t in beam_tokens],\
               [self.decoder.detokenize(t) for t in diff_tokens],\
               post_corrected_pep, post_positional_score,\
               diff_scores, aa_beam_scores


    def beam_post_correction(self, matrix, mass, c, ppm=10):
        positional_score, seq = torch.max(matrix, axis = -1) #np.max(matrix, axis=-1)
        pep = self.decoder.detokenize(seq)
        if '$' in pep: pep.remove('$')
        seq = seq[:len(pep)+1]
        tol = mass * ppm / 1000000
        if len(pep) < 1:
            return '', -1, positional_score
        mass_list = list(self.peptide_mass_calculator.masses.values())
        msp = self.peptide_mass_calculator.mass(pep, c)
        #print('charge, mass, predicted mass', c, mass, msp)
        delta = msp - mass
        pos = 0
        a = seq[0]

        if abs(delta) < tol:
            #return self.topep(seq), -1, positional_score
            return self.decoder.detokenize(seq), -1, positional_score

        for i in range(len(seq) - 1):  # no last pos
            mi = mass_list[seq[i]]
            for j in range(self.decoder.vocab_size -1):


                d = msp - mass + (mass_list[j] - mi) / c

                if abs(d) < abs(delta):
                    delta = d
                    pos = i
                    a = j

        if abs(delta) < tol:  # have good match
            candi = (seq == seq[pos]).int()
            if torch.sum(candi) > 1.5:  # ambiguis
                max_prob, pos_seq = torch.max(matrix[:len(seq)], axis=-1)
                pos = torch.argmin((1-c)*10 + c*max_prob)

            seq[pos] = a
            positional_score[pos] = 1
            return self.decoder.detokenize(seq), pos, positional_score
        else:
            return self.decoder.detokenize(seq), -1, positional_score

    def beam_search_decode(
            self, spectra: torch.Tensor, ms1: torch.Tensor, ms2_b_ion: torch.Tensor, ms2_y_ion: torch.Tensor, precursors: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Beam search decoding of the spectrum predictions.

        Return the highest scoring peptide, within the precursor m/z tolerance
        whenever possible.

        Parameters
        ----------
       spectra : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra for which to predict peptide sequences.
            Axis 0 represents an MS/MS spectrum, axis 1 contains the peaks in
            the MS/MS spectrum, and axis 2 is essentially a 2-tuple specifying
            the m/z-intensity pair for each peak. These should be zero-padded,
            such that all of the spectra in the batch are the same length.
        ms1 : torch.Tensor of shape (n_spectra, n_neighbors)
             the elusion profile for each precursor ion
        ms2_b_ion : torch.Tensor of shape (n_spectra, seq_len, vocab_size,n_ion_type, n_neighbors, window_size)
                  prodcut b ion peptide
        ms2_y_ion : torch.Tensor of shape (n_spectra, seq_len, vocab_size,n_ion_type, n_neighbors, window_size)
                   product y ion peptide
        precursors : torch.Tensor of size (n_spectra, 3)
            The measured precursor mass (axis 0), precursor charge (axis 1), and
            precursor m/z (axis 2) of each MS/MS spectrum.
        Returns
        -------
        scores : torch.Tensor of shape (n_spectra, max_length, n_amino_acids)
            The individual amino acid scores for each prediction.
        tokens : torch.Tensor of shape (n_spectra, max_length)
            The predicted tokens for each spectrum.
        """
        memories, mem_masks = self.encoder(spectra, ms1, ms2_b_ion, ms2_y_ion)

        # Sizes.
        batch = spectra.shape[0]  # B
        length = self.max_length + 1  # L
        vocab = self.decoder.vocab_size + 1  # V
        beam = self.n_beams  # S

        # Initialize scores and tokens.
        scores = torch.full(
            size=(batch, length, vocab, beam), fill_value=torch.nan
        )
        scores = scores.type_as(spectra)
        tokens = torch.zeros(batch, length, beam, dtype=torch.int64)
        tokens = tokens.to(self.encoder.device)
        # Keep track whether terminated beams have fitting precursor m/z.
        beam_fits_prec_tol = torch.zeros(batch * beam, dtype=torch.bool)

        # Create cache for decoded beams.
        (
            cache_scores,
            cache_tokens,
            cache_next_idx,
            cache_pred_seq,
            cache_pred_score,
        ) = self._create_beamsearch_cache(scores, tokens)

        # Get the first prediction.
        pred, _ = self.decoder(None,None,  precursors, memories, mem_masks)
        tokens[:, 0, :] = torch.topk(pred[:, 0, :], beam, dim=1)[1]
        scores[:, :1, :, :] = einops.repeat(pred, "B L V -> B L V S", S=beam)

        # Make all tensors the right shape for decoding.
        precursors = einops.repeat(precursors, "B L -> (B S) L", S=beam)
        mem_masks = einops.repeat(mem_masks, "B L -> (B S) L", S=beam)
        memories = einops.repeat(memories, "B L V -> (B S) L V", S=beam)
        scores = einops.rearrange(scores, "B L V S -> (B S) L V")
        tokens = einops.rearrange(tokens, "B L S -> (B S) L")

        # The main decoding loop.
        for i in range(1, self.max_length + 1):
            # Terminate beams exceeding precursor m/z tolerance and track all
            # terminated beams.
            finished_beams_idx, tokens = self._terminate_finished_beams(
                tokens, precursors, beam_fits_prec_tol, i
            )
            # Cache terminated beams, group and order by fitting precursor m/z
            # and confidence score.
            self._cache_finished_beams(
                finished_beams_idx,
                cache_next_idx,
                cache_pred_seq,
                cache_pred_score,
                cache_tokens,
                cache_scores,
                tokens,
                scores,
                beam_fits_prec_tol,
                i,
            )
            # Reset precursor tolerance status of all beams.
            beam_fits_prec_tol = torch.zeros(batch * beam, dtype=torch.bool)

            # Stop decoding when all current beams are terminated.
            decoded = (tokens == self.stop_token).any(axis=1)
            if decoded.all():
                break
            # Update the scores.
            scores[~decoded, : i + 1, :], _ = self.decoder(
                tokens[~decoded, :i],
                None, 
                precursors[~decoded, :],
                memories[~decoded, :, :],
                mem_masks[~decoded, :],
            )
            # Find top-k beams with highest scores and continue decoding those.
            scores, tokens = self._get_topk_beams(scores, tokens, batch, i)

        # Return the peptide with the highest confidence score, within the
        # precursor m/z tolerance if possible.
        output_tokens, output_scores = self._get_top_peptide(
            cache_pred_score, cache_tokens, cache_scores, batch
        )
        return self.softmax(output_scores), output_tokens

    def _create_beamsearch_cache(
        self, scores: torch.Tensor, tokens: torch.Tensor
    ) -> Tuple[
        torch.Tensor,
        torch.Tensor,
        Dict[int, int],
        Dict[int, Set[str]],
        Dict[int, List[List[Tuple[float, int]]]],
    ]:
        """
        Create cache tensor and dictionary to store and group terminated beams.

        Parameters
        ----------
        scores : torch.Tensor of shape
        (n_spectra, max_length, n_amino_acids, n_beams)
            Output scores of the model.
        tokens : torch.Tensor of size (n_spectra, max_length, n_beams)
            Output token of the model corresponding to amino acid sequences.

        Returns
        -------
        cache_scores : torch.Tensor of shape
        (n_spectra * n_beams, max_length, n_amino_acids)
            The score for each amino acid in cached peptides.
        cache_tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            The token for each amino acid in cached peptides.
        cache_next_idx : Dict[int, int]
            Next available tensor index to cache peptides for each spectrum.
        cache_pred_seq : Dict[int, Set[torch.Tensor]]
            Set of decoded peptide tokens for each spectrum.
        cache_pred_score : Dict[int, List[List[Tuple[float, int]]]
            Confidence score for each decoded peptide, separated as
            precursor m/z fitting vs not, for each spectrum.
        """
        batch, beam = scores.shape[0], scores.shape[-1]

        # Cache terminated beams and their scores.
        cache_scores = einops.rearrange(scores.clone(), "B L V S -> (B S) L V")
        cache_tokens = einops.rearrange(tokens.clone(), "B L S -> (B S) L")

        # Keep pointer to free rows in the cache and already cached predictions.
        cache_next_idx = {i: i * beam for i in range(batch)}
        # Keep already decoded peptides to avoid duplicates in cache.
        cache_pred_seq = {i: set() for i in range(batch)}
        # Store peptide scores to replace lower score peptides in cache with
        # higher score peptides during decoding.
        cache_pred_score = {i: [[], []] for i in range(batch)}

        return (
            cache_scores,
            cache_tokens,
            cache_next_idx,
            cache_pred_seq,
            cache_pred_score,
        )

    def _terminate_finished_beams(
        self,
        tokens: torch.Tensor,
        precursors: torch.Tensor,
        is_beam_prec_fit: torch.Tensor,
        idx: int,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Terminate beams exceeding the precursor m/z tolerance.

        Track all terminated beams.

        Parameters
        ----------
        tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            Output token of the model corresponding to amino acid sequences.
        precursors : torch.Tensor of size (n_spectra * n_beams, 3)
            The measured precursor mass (axis 0), precursor charge (axis 1), and
            precursor m/z (axis 2) of each MS/MS spectrum.
        is_beam_prec_fit: torch.Tensor of shape (n_spectra * n_beams)
            Boolean tensor indicating if current beams are within precursor m/z
            tolerance.
        idx : int
            Index to be considered in the current decoding step.

        Returns
        -------
        finished_beams_idx : torch.Tensor
            Indices of all finished beams on tokens tensor.
        tokens : torch.Tensor of size (n_spectra * n_beams, max_length)
            Output token of the model corresponding to amino acid sequences.
        """
        # Check for tokens with a negative mass (i.e. neutral loss).
        aa_neg_mass = [None]
        for aa, mass in self.peptide_mass_calculator.masses.items():
            if mass < 0:
                aa_neg_mass.append(aa)

        # Terminate beams that exceed the precursor m/z.
        for beam_i in range(len(tokens)):
            # Check only non-terminated beams.
            if self.stop_token not in tokens[beam_i]:
                # Finish if dummy was predicted at the previous step.
                if tokens[beam_i][idx - 1] == 0:
                    tokens[beam_i][idx - 1] = self.stop_token
                # Terminate the beam if it exceeds the precursor m/z tolerance.
                else:
                    precursor_charge = precursors[beam_i, 1].item()
                    precursor_mz = precursors[beam_i, 2].item()
                    # Only terminate if the m/z difference cannot be corrected
                    # anymore by a subsequently predicted AA with negative mass.
                    matches_precursor_mz = exceeds_precursor_mz = False
                    for aa in aa_neg_mass:
                        peptide = self.decoder.detokenize(tokens[beam_i][:idx])
                        if aa is not None:
                            peptide.append(aa)
                        try:
                            calc_mz = self.peptide_mass_calculator.mass(
                                seq=peptide, charge=precursor_charge
                            )
                            delta_mass_ppm = [
                                _calc_mass_error(
                                    calc_mz,
                                    precursor_mz,
                                    precursor_charge,
                                    isotope,
                                )
                                for isotope in range(
                                    self.isotope_error_range[0],
                                    self.isotope_error_range[1] + 1,
                                )
                            ]
                            # Terminate the beam if the calculated m/z for the
                            # predicted peptide (without potential additional
                            # AAs with negative mass) is within the precursor
                            # m/z tolerance.
                            matches_precursor_mz = aa is None and any(
                                abs(d) < self.precursor_mass_tol
                                for d in delta_mass_ppm
                            )
                            # Terminate the beam if the calculated m/z exceeds
                            # the precursor m/z + tolerance and hasn't been
                            # corrected by a subsequently predicted AA with
                            # negative mass.
                            exceeds_precursor_mz = aa is not None and all(
                                d > self.precursor_mass_tol
                                for d in delta_mass_ppm
                            )
                            if matches_precursor_mz or exceeds_precursor_mz:
                                break
                        except KeyError:
                            matches_precursor_mz = exceeds_precursor_mz = False
                    if matches_precursor_mz or exceeds_precursor_mz:
                        tokens[beam_i][idx] = self.stop_token
                        is_beam_prec_fit[beam_i] = matches_precursor_mz

        # Get the indices of finished beams.
        finished_idx = torch.where((tokens == self.stop_token).any(dim=1))[0]
        return finished_idx, tokens

    def _cache_finished_beams(
        self,
        finished_beams_idx: torch.Tensor,
        cache_next_idx: Dict[int, int],
        cache_pred_seq: Dict[int, Set[torch.Tensor]],
        cache_pred_score: Dict[int, List[List[Tuple[float, int]]]],
        cache_tokens: torch.Tensor,
        cache_scores: torch.Tensor,
        tokens: torch.Tensor,
        scores: torch.Tensor,
        is_beam_prec_fit: torch.Tensor,
        idx: int,
    ):
        """
        Cache terminated beams.

        Group and order by fitting precursor m/z and confidence score.

        Parameters
        ----------
        finished_beams_idx : torch.Tensor
            Indices of all finished beams on tokens tensor.
        cache_next_idx : Dict[int, int]
            Next available tensor index to cache peptides for each spectrum.
        cache_pred_seq : Dict[int, Set[torch.Tensor]]
            Set of decoded peptide tokens for each spectrum.
        cache_pred_score : Dict[int, List[List[Tuple[float, int]]]
            Confidence score for each decoded peptide, separated as
            precursor m/z fitting vs not, for each spectrum.
        cache_tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            The token for each amino acid in cached peptides.
        cache_scores : torch.Tensor of shape
        (n_spectra * n_beams, max_length, n_amino_acids)
            The score for each amino acid in cached peptides.
        tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            Output token of the model corresponding to amino acid sequences.
         scores : torch.Tensor of shape
         (n_spectra *  n_beams, max_length, n_amino_acids)
            Output scores of the model.
        is_beam_prec_fit: torch.Tensor of shape (n_spectra * n_beams)
            Boolean tensor indicating if current beams are within the precursor
            m/z tolerance.
        idx : int
            Index to be considered in the current decoding step.
        """
        beam = self.n_beams
        # Store finished beams in the cache.
        for i in finished_beams_idx:
            i = i.item()
            spec_idx = i // beam  # Find the starting index of the spectrum.
            # Check position of stop token (changes in case stopped early).
            stop_token_idx = idx - (not tokens[i][idx] == self.stop_token)
            # Check if predicted peptide already in cache.
            pred_seq = tokens[i][:stop_token_idx]
            is_peptide_cached = any(
                torch.equal(pep, pred_seq) for pep in cache_pred_seq[spec_idx]
            )
            # Don't cache this peptide if it was already predicted previously.
            if is_peptide_cached:
                continue
            smx = self.softmax(scores)
            aa_scores = [smx[i, j, k].item() for j, k in enumerate(pred_seq)]
            pep_score = _aa_to_pep_score(aa_scores)
            # Cache peptides with fitting (idx=0) or non-fitting (idx=1)
            # precursor m/z separately.
            cache_pred_score_idx = cache_pred_score[spec_idx]
            cache_i = int(not is_beam_prec_fit[i])
            # Directly cache if we don't already have k peptides cached.
            if cache_next_idx[spec_idx] < (spec_idx + 1) * beam:
                insert_idx = cache_next_idx[spec_idx]
                cache_next_idx[spec_idx] += 1  # Move the pointer.
                heap_update = heapq.heappush
            # If any prediction has a non-fitting precursor m/z and this
            # prediction has a fitting precursor m/z, replace the non-fitting
            # peptide with the lowest score, irrespective of the current
            # predicted score.
            elif is_beam_prec_fit[i] and len(cache_pred_score_idx[1]) > 0:
                _, insert_idx = heapq.heappop(cache_pred_score_idx[1])
                heap_update = heapq.heappush
            # Else, replace the lowest-scoring peptide with corresponding
            # fitting or non-fitting precursor m/z if the current predicted
            # score is higher.
            elif len(cache_pred_score_idx[cache_i]) > 0:
                # Peek at the top of the heap (lowest score).
                pop_pep_score, insert_idx = cache_pred_score_idx[cache_i][0]
                heap_update = heapq.heappushpop
                # Don't store this prediction if it has a lower score than all
                # previous predictions.
                if pep_score <= pop_pep_score:
                    continue
            # Finally, no matching cache found (we should never get here).
            else:
                continue
            # Store the current prediction in its relevant cache.
            cache_tokens[insert_idx, :] = tokens[i, :]
            cache_scores[insert_idx, :, :] = scores[i, :, :]
            heap_update(cache_pred_score_idx[cache_i], (pep_score, insert_idx))
            cache_pred_seq[spec_idx].add(pred_seq)

    def _get_top_peptide(
        self,
        cache_pred_score: Dict[int, List[List[Tuple[float, int]]]],
        cache_tokens: torch.tensor,
        cache_scores: torch.tensor,
        batch: int,
    ) -> Tuple[torch.tensor, torch.tensor]:
        """
        Return the peptide with the highest confidence score for each spectrum.

        If there are no peptides within the precursor m/z tolerance, return the
        highest-scoring peptide among the non-fitting predictions.

        Parameters
        ----------
        cache_pred_score : Dict[int, List[List[Tuple[float, int]]]
            Confidence score for each decoded peptide, separated as
            precursor m/z fitting vs not, for each spectrum.
        cache_tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            The token for each amino acid in cached peptides.
        cache_scores : torch.Tensor of shape
        (n_spectra * n_beams, max_length, n_amino_acids)
            The score for each amino acid in cached peptides.
        batch: int
            Number of spectra in the batch.

        Returns
        -------
        output_tokens : torch.Tensor of shape (n_spectra, max_length)
            The token for each amino acid in the output peptides.
        output_scores : torch.Tensor of shape
        (n_spectra, max_length, n_amino_acids)
            The score for each amino acid in cached peptides.
        """
        # Sizes.
        length = self.max_length + 1  # L
        vocab = self.decoder.vocab_size + 1  # V

        # Create output tensors for top scoring peptides and their scores.
        output_scores = torch.full(
            size=(batch, length, vocab), fill_value=torch.nan
        )
        output_scores = output_scores.type_as(cache_scores)
        output_tokens = torch.zeros(batch, length).type_as(cache_tokens)

        # Return the top scoring peptide (fitting precursor mass if possible).
        for spec_idx in range(batch):
            cache = cache_pred_score[spec_idx][
                len(cache_pred_score[spec_idx][0]) == 0
            ]
            # Skip this spectrum if it doesn't have any finished beams.
            if len(cache) == 0:
                continue
            _, top_score_idx = max(cache, key=operator.itemgetter(0))
            output_tokens[spec_idx, :] = cache_tokens[top_score_idx, :]
            output_scores[spec_idx, :, :] = cache_scores[top_score_idx, :, :]

        return output_tokens, output_scores

    def _get_topk_beams(
        self, scores: torch.tensor, tokens: torch.tensor, batch: int, idx: int
    ) -> Tuple[torch.tensor, torch.tensor]:
        """
        Find top-k beams with highest confidences and continue decoding those.

        Discontinue decoding for beams where the stop token was predicted.

        Parameters
        ----------
        scores : torch.Tensor of shape
        (n_spectra * n_beams, max_length, n_amino_acids)
            Output scores of the model.
        tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            Output token of the model corresponding to amino acid sequences.
        batch: int
            Number of spectra in the batch.
        idx : int
            Index to be considered in the current decoding step.

        Returns
        -------
        scores : torch.Tensor of shape
        (n_spectra * n_beams, max_length, n_amino_acids)
            Output scores of the model.
        tokens : torch.Tensor of shape (n_spectra * n_beams, max_length)
            Output token of the model corresponding to amino acid sequences.
        """
        beam = self.n_beams  # S
        vocab = self.decoder.vocab_size + 1  # V

        # Reshape to group by spectrum (B for "batch").
        scores = einops.rearrange(scores, "(B S) L V -> B L V S", S=beam)
        tokens = einops.rearrange(tokens, "(B S) L -> B L S", S=beam)
        prev_tokens = einops.repeat(
            tokens[:, :idx, :], "B L S -> B L V S", V=vocab
        )

        # Get the previous tokens and scores.
        prev_scores = torch.gather(
            scores[:, :idx, :, :], dim=2, index=prev_tokens
        )
        prev_scores = einops.repeat(
            prev_scores[:, :, 0, :], "B L S -> B L (V S)", V=vocab
        )

        # Get scores for all possible beams at this step.
        step_scores = torch.zeros(batch, idx + 1, beam * vocab).type_as(scores)
        step_scores[:, :idx, :] = prev_scores
        step_scores[:, idx, :] = einops.rearrange(
            scores[:, idx, :, :], "B V S -> B (V S)"
        )

        # Mask out terminated beams. Include delta mass induced termination.
        extended_prev_tokens = einops.repeat(
            tokens[:, : idx + 1, :], "B L S -> B L V S", V=vocab
        )
        finished_mask = (
            einops.rearrange(extended_prev_tokens, "B L V S -> B L (V S)")
            == self.stop_token
        ).any(axis=1)
        # Mask out the index '0', i.e. padding token, by default.
        finished_mask[:, :beam] = True

        # Figure out the top K decodings.
        _, top_idx = torch.topk(
            step_scores.nanmean(dim=1) * (~finished_mask).float(), beam
        )
        v_idx, s_idx = np.unravel_index(top_idx.cpu(), (vocab, beam))
        s_idx = einops.rearrange(s_idx, "B S -> (B S)")
        b_idx = einops.repeat(torch.arange(batch), "B -> (B S)", S=beam)

        # Record the top K decodings.
        tokens[:, :idx, :] = einops.rearrange(
            prev_tokens[b_idx, :, 0, s_idx], "(B S) L -> B L S", S=beam
        )
        tokens[:, idx, :] = torch.tensor(v_idx)
        scores[:, : idx + 1, :, :] = einops.rearrange(
            scores[b_idx, : idx + 1, :, s_idx], "(B S) L V -> B L V S", S=beam
        )
        scores = einops.rearrange(scores, "B L V S -> (B S) L V")
        tokens = einops.rearrange(tokens, "B L S -> (B S) L")
        return scores, tokens

    def decode_q_sample(self, seq2d):
        return torch.argmax(seq2d, dim=2).int()
    
    def _forward_step(
        self,
        spectra: torch.Tensor,
        ms1: torch.Tensor,
        ms2_b_ion: torch.Tensor,
        ms2_y_ion: torch.Tensor, 
        precursors: torch.Tensor,
        sequences: List[str],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        The forward learning step.

        Parameters
        ----------
        spectra : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra for which to predict peptide sequences.
            Axis 0 represents an MS/MS spectrum, axis 1 contains the peaks in
            the MS/MS spectrum, and axis 2 is essentially a 2-tuple specifying
            the m/z-intensity pair for each peak. These should be zero-padded,
            such that all of the spectra in the batch are the same length.
        ms1: torch.Tensor of shape (n_spectra, n_neighbors)
             the elusion profile for each precursor ion
        ms2_b_ion: torch.Tensor of shape (n_spectra, seq_len, vocab_size,n_ion_type, n_neighbors, window_size)
                  prodcut b ion peptide
        ms2_y_ion: torch.Tensor of shape (n_spectra, seq_len, vocab_size,n_ion_type, n_neighbors, window_size)
                   product y ion peptide
        precursors : torch.Tensor of size (n_spectra, 3)
            The measured precursor mass (axis 0), precursor charge (axis 1), and
            precursor m/z (axis 2) of each MS/MS spectrum.
        sequences : List[str] of length n_spectra
            The partial peptide sequences to predict.

        Returns
        -------
        scores : torch.Tensor of shape (n_spectra, length, n_amino_acids)
            The individual amino acid scores for each prediction.
        tokens : torch.Tensor of shape (n_spectra, length)
            The predicted tokens for each spectrum.
        """
        batch_size = spectra.shape[0]

        time = torch.randint(0, self.time_steps,(batch_size,)).long().to(spectra.device)
        if sequences is not None:
             sequences = depthcharge.utils.listify(sequences)
             tokens = [self.decoder.tokenize(s) for s in sequences]
             tokens = torch.nn.utils.rnn.pad_sequence(tokens, batch_first=True)
        #adding noise to each amino acid token
        q_sample = self.forward_diffusion(tokens)
        q_tokens = self.decode_q_sample(q_sample)
        sequences = [''.join(self.decoder.detokenize(token)) for token in q_tokens]

        pred, truth = self.decoder(sequences, time, precursors, *self.encoder(spectra, ms1, ms2_b_ion, ms2_y_ion))
        return pred, truth

    def training_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, List[str]],
        *args,
        mode: str = "train",
    ) -> torch.Tensor:
        """
        A single training step.

        Parameters
        ----------
        batch : Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, List[str]]
            A batch of (i) MS/MS spectra, (ii) ms1, (iii) ms2_b_ion, (iv)ms2_y_ion  (v) precursor information, (vi)
            peptide sequences as torch Tensors.
        mode : str
            Logging key to describe the current stage.
        Returns
        -------
        torch.Tensor
            The loss of the training step.
        """
        pred, truth = self._forward_step(*batch)
        pred = pred[:, :-1, :].reshape(-1, self.decoder.vocab_size + 1)
        loss = self.celoss(pred, truth.flatten())
        self.log(
            "CELoss",
            {mode: loss.detach()},
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        return loss

    def process_pep(self, peptide):
        if isinstance(peptide, str):
            peptide = re.split(r"(?<=.)(?=[A-Z])", peptide)
            peptide = ''.join(peptide)
        return peptide

    def validation_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, List[str]], *args
    ) -> torch.Tensor:
        """
        A single validation step.

        Parameters
        ----------
        batch : Tuple[torch.Tensor, torch.Tensor, List[str]]
            A batch of (i) MS/MS spectra, (ii) ms1, (iii) ms2_b_ion, (iv)ms2_y_ion, (v) precursor information, (vi)
            peptide sequences.

        Returns
        -------
        torch.Tensor
            The loss of the validation step.
        """
        # Record the loss.
        loss = self.training_step(batch, mode="valid")

        # Calculate and log amino acid and peptide match evaluation metrics from
        # the predicted peptides.
        beam_pred_peps, diff_pred_peps, post_peptides, post_pep_scores, diff_aa_scores, beam_aa_scores = self.forward(batch[0], batch[1], batch[2], batch[3], batch[4])
        # FIXME: Temporary fix to skip predictions with multiple stop tokens.
        peptides_pred, peptides_true = [], []
        for beam_pred_pep, peptide_true, diff_pred_pep, post_pep, post_score_pep,beam_aa_score, diff_aa_score, precursor in zip(beam_pred_peps, batch[5], diff_pred_peps,post_peptides, post_pep_scores, beam_aa_scores, diff_aa_scores, batch[1]):
            if len(beam_pred_pep) > 0:
                if beam_pred_pep[0] == "$":
                    beam_pred_pep = beam_pred_pep[1:]  # Remove stop token.
                if "$" not in beam_pred_pep and len(beam_pred_pep) > 0:
                    self.psm[peptide_true] = ''.join(post_pep)

                    peptides_pred.append(beam_pred_pep)
                    peptides_true.append(peptide_true)
                    peptide,aa_tokens,peptide_score,aa_scores = self._get_output_peptide_and_scores(
                        beam_pred_pep, beam_aa_score
                    )
                    peptide_pred = ''.join(beam_pred_pep)
                    if self.out_writer:
                        self.out_writer.psms.append(
                        (
                            beam_pred_pep,
                            peptide_true,
                            peptide_score,
                            0,
                            0,
                            0,
                            aa_scores,
                        ),
                    )

        aa_precision, aa_recall, pep_recall = evaluate.aa_match_metrics(
            *evaluate.aa_match_batch(
                peptides_pred, peptides_true, self.decoder._peptide_mass.masses
            )
        )


        log_args = dict(on_step=False, on_epoch=True, sync_dist=True)
        self.log("aa_precision", {"valid": aa_precision}, **log_args)
        self.log("aa_recall", {"valid": aa_recall}, **log_args)
        self.log("pep_recall", {"valid": pep_recall}, **log_args)

        return loss, peptides_true, peptides_pred

    def predict_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, str], *args
    ) -> Tuple[torch.Tensor, torch.Tensor, List[List[str]], torch.Tensor]:
        """
        A single prediction step.

        Parameters
        ----------
        batch : Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
            A batch of (i) MS/MS spectra, (ii) ms1 (elution profile),
             (iii) ms2_b_ion product ion peptide (v) ms2_y_ion (iv)precursor information, (vi) spectrum identifiers as torch Tensors.

        Returns
        -------
        spectrum_idx : torch.Tensor
            The spectrum identifiers.
        precursors : torch.Tensor
            Precursor information for each spectrum.
        peptides : List[List[str]]
            The predicted peptide sequences for each spectrum.
        aa_scores : torch.Tensor of shape (n_spectra, length, n_amino_acids)
            The individual amino acid scores for each prediction.
        """

        peptides, aa_scores = self.forward(batch[0], batch[1], batch[2], batch[3], batch[4])
        return batch[5], batch[4], peptides, aa_scores

    def on_train_epoch_end(self) -> None:
        """
        Log the training loss at the end of each epoch.
        """
        train_loss = self.trainer.callback_metrics["CELoss"]["train"].detach()
        self._history[-1]["train"] = train_loss
        self._log_history()

    def on_validation_epoch_end(self) -> None:
        """
        Log the validation metrics at the end of each epoch.
        """
        callback_metrics = self.trainer.callback_metrics
        metrics = {
            "epoch": self.trainer.current_epoch,
            "valid": callback_metrics["CELoss"]["valid"].detach(),
            "valid_aa_precision": callback_metrics["aa_precision"][
                "valid"
            ].detach(),
            "valid_aa_recall": callback_metrics["aa_recall"]["valid"].detach(),
            "valid_pep_recall": callback_metrics["pep_recall"][
                "valid"
            ].detach(),
        }
        self._history.append(metrics)
        self._log_history()

    def on_predict_epoch_end(
        self, results: List[List[Tuple[np.ndarray, List[str], torch.Tensor]]]
    ) -> None:
        """
        Write the predicted peptide sequences and amino acid scores to the
        output file.
        """
        if self.out_writer is None:
            return
        for batch in results:
            for step in batch:
                for spectrum_i, precursor, aa_tokens, aa_scores in zip(*step):
                    # Get peptide sequence, amino acid and peptide-level
                    # confidence scores to write to output file.
                    (
                        peptide,
                        aa_tokens,
                        peptide_score,
                        aa_scores,
                    ) = self._get_output_peptide_and_scores(
                        aa_tokens, aa_scores
                    )
                    # Compare the experimental vs calculated precursor m/z.
                    _, precursor_charge, precursor_mz = precursor
                    precursor_charge = int(precursor_charge.item())
                    precursor_mz = precursor_mz.item()
                    try:
                        calc_mz = self.peptide_mass_calculator.mass(
                            aa_tokens, precursor_charge
                        )
                        delta_mass_ppm = [
                            _calc_mass_error(
                                calc_mz,
                                precursor_mz,
                                precursor_charge,
                                isotope,
                            )
                            for isotope in range(
                                self.isotope_error_range[0],
                                self.isotope_error_range[1] + 1,
                            )
                        ]
                        is_within_precursor_mz_tol = any(
                            abs(d) < self.precursor_mass_tol
                            for d in delta_mass_ppm
                        )
                    except KeyError:
                        calc_mz, is_within_precursor_mz_tol = np.nan, False
                    # Subtract one if the precursor m/z tolerance is violated.
                    if not is_within_precursor_mz_tol:
                        peptide_score -= 1

                    self.out_writer.psms.append(
                        (
                            peptide,
                            spectrum_i,
                            peptide_score,
                            precursor_charge,
                            precursor_mz,
                            calc_mz,
                            aa_scores,
                        ),
                    )

    def _get_output_peptide_and_scores(
        self, aa_tokens: List[str], aa_scores: torch.Tensor
    ) -> Tuple[str, List[str], float, str]:
        """
        Get peptide to output, amino acid and peptide-level confidence scores.

        Parameters
        ----------
        aa_tokens : List[str]
            Amino acid tokens of the peptide sequence.
        aa_scores : torch.Tensor
            Amino acid-level confidence scores for the predicted sequence.

        Returns
        -------
        peptide : str
            Peptide sequence.
        aa_tokens : List[str]
            Amino acid tokens of the peptide sequence.
        peptide_score : str
            Peptide-level confidence score.
        aa_scores : str
            Amino acid-level confidence scores for the predicted sequence.
        """
        # Omit stop token.
        aa_tokens = aa_tokens[1:] if self.decoder.reverse else aa_tokens[:-1]
        peptide = "".join(aa_tokens)

        # If this is a non-finished beam (after exceeding `max_length`), return
        # a dummy (empty) peptide and NaN scores.
        if len(peptide) == 0:
            aa_tokens = []

        # Take scores corresponding to the predicted amino acids. Reverse tokens
        # to correspond with correct amino acids as needed.
        step = -1 if self.decoder.reverse else 1
        top_aa_scores = [
            aa_score[self.decoder._aa2idx[aa_token]].item()
            for aa_score, aa_token in zip(aa_scores, aa_tokens[::step])
        ][::step]

        # Get peptide-level score from amino acid-level scores.
        peptide_score = _aa_to_pep_score(top_aa_scores)
        aa_scores = ",".join(list(map("{:.5f}".format, top_aa_scores)))
        return peptide, aa_tokens, peptide_score, aa_scores

    def _log_history(self) -> None:
        """
        Write log to console, if requested.
        """
        # Log only if all output for the current epoch is recorded.
        if len(self._history) > 0 and len(self._history[-1]) == 6:
            if len(self._history) == 1:
                logger.info(
                    "Epoch\tTrain loss\tValid loss\tAA precision\tAA recall\t"
                    "Peptide recall"
                )
            metrics = self._history[-1]
            if metrics["epoch"] % self.n_log == 0:
                logger.info(
                    "%i\t%.6f\t%.6f\t%.6f\t%.6f\t%.6f",
                    metrics["epoch"] + 1,
                    metrics.get("train", np.nan),
                    metrics.get("valid", np.nan),
                    metrics.get("valid_aa_precision", np.nan),
                    metrics.get("valid_aa_recall", np.nan),
                    metrics.get("valid_pep_recall", np.nan),
                )
                if self.tb_summarywriter is not None:
                    for descr, key in [
                        ("loss/train_crossentropy_loss", "train"),
                        ("loss/dev_crossentropy_loss", "valid"),
                        ("eval/dev_aa_precision", "valid_aa_precision"),
                        ("eval/dev_aa_recall", "valid_aa_recall"),
                        ("eval/dev_pep_recall", "valid_pep_recall"),
                    ]:
                        self.tb_summarywriter.add_scalar(
                            descr,
                            metrics.get(key, np.nan),
                            metrics["epoch"] + 1,
                        )

    def configure_optimizers(
        self,
    ) -> Tuple[torch.optim.Optimizer, Dict[str, Any]]:
        """
        Initialize the optimizer.

        This is used by pytorch-lightning when preparing the model for training.

        Returns
        -------
        Tuple[torch.optim.Optimizer, Dict[str, Any]]
            The initialized Adam optimizer and its learning rate scheduler.
        """
        optimizer = torch.optim.Adam(self.parameters(), **self.opt_kwargs)
        # Apply learning rate scheduler per step.
        lr_scheduler = CosineWarmupScheduler(
            optimizer, warmup=self.warmup_iters, max_iters=self.max_iters
        )
        return [optimizer], {"scheduler": lr_scheduler, "interval": "step"}


class CosineWarmupScheduler(torch.optim.lr_scheduler._LRScheduler):
    """
    Learning rate scheduler with linear warm up followed by cosine shaped decay.

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Optimizer object.
    warmup : int
        The number of warm up iterations.
    max_iters : torch.optim
        The total number of iterations.
    """

    def __init__(
        self, optimizer: torch.optim.Optimizer, warmup: int, max_iters: int
    ):
        self.warmup, self.max_iters = warmup, max_iters
        super().__init__(optimizer)

    def get_lr(self):
        lr_factor = self.get_lr_factor(epoch=self.last_epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        lr_factor = 0.5 * (1 + np.cos(np.pi * epoch / self.max_iters))
        if epoch <= self.warmup:
            lr_factor *= epoch / self.warmup
        return lr_factor


def _calc_mass_error(
    calc_mz: float, obs_mz: float, charge: int, isotope: int = 0
) -> float:
    """
    Calculate the mass error in ppm between the theoretical m/z and the observed
    m/z, optionally accounting for an isotopologue mismatch.

    Parameters
    ----------
    calc_mz : float
        The theoretical m/z.
    obs_mz : float
        The observed m/z.
    charge : int
        The charge.
    isotope : int
        Correct for the given number of C13 isotopes (default: 0).

    Returns
    -------
    float
        The mass error in ppm.
    """
    return (calc_mz - (obs_mz - isotope * 1.00335 / charge)) / obs_mz * 10**6


def _aa_to_pep_score(aa_scores: List[float]) -> float:
    """
    Calculate peptide-level confidence score from amino acid level scores.

    Parameters
    ----------
    aa_scores : List[float]
        Amino acid level confidence scores.

    Returns
    -------
    float
        Peptide confidence score.
    """
    return np.mean(aa_scores)



"""Preprocessing functions for mass spectra.

These functions can be used with datasets (`SpectrumDataset` and
`AnnotatedSpectrumDataset`) or our `SpectrumDataModule`.

One or more preprocessing function can be applied to each mass spectrum during
batch loading using the `preprocessing_fn` parameter of these classes. To apply
preprocessing steps sequentially, merely pass a list of functions to this
argument.

We can also define custom preprocessing functions. All preprocessing functions
must accept 4 keyword arguments:
    - `mz_array` : torch.Tensor of shape (n_peaks,)
        The m/z values of the peaks in the spectrum.
    - `int_array` : torch.Tensor of shape (n_peaks,)
        The intensity values of the peaks in the spectrum.
    - `precursor_mz` : float
        The precursor m/z.
    - `precursor_charge` : int
        The precursor charge.

Preprocessing functions always return the processed `mz_array` and `int_array`.


### Examples

Remove the peaks around the precursor m/z then square root transform
intensities and scale to unit norm:
```Python
SpectrumDataset(
    ...,
    preprocessing_fn=[
        preprocessing.remove_precursor_preak,
        preprocessing.sqrt_and_norm,
    ],
)
```

Apply a custom function:
```
def my_func(mz_array, int_array, precursor_mz, precursor_charge):
    return mz_array, torch.log(int_array)

SpectrumDataset(
    ...,
    preprocessing_fn=my_func,
)
```

"""
from typing import Tuple

import torch


def sqrt_and_norm(
    mz_array: torch.Tensor,
    int_array: torch.Tensor,
    precursor_mz: float,
    precursor_charge: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Square root the intensities and scale to unit norm.

    Parameters
    ----------
    mz_array : torch.Tensor of shape (n_peaks,)
        The m/z values of the peaks in the spectrum.
    int_array : torch.Tensor of shape (n_peaks,)
        The intensity values of the peaks in the spectrum.
    precursor_mz : float
        The precursor m/z.
    precursor_charge : int
        The precursor charge.

    Returns
    -------
    mz_array : torch.Tensor of shape (n_peaks,)
        The m/z values of the peaks in the spectrum.
    int_array : torch.Tensor of shape (n_peaks,)
        The intensity values of the peaks in the spectrum.
    """
    int_array = torch.sqrt(int_array)
    int_array /= torch.linalg.norm(int_array)
    return mz_array, int_array


def remove_precursor_peak(
    mz_array: torch.Tensor,
    int_array: torch.Tensor,
    precursor_mz: float,
    tol: float = 1.5,
) -> Tuple[torch.Tensor, torch.Tensor]:

    """Square root the intensities and scale to unit norm.

    Parameters
    ----------
    mz_array : torch.Tensor of shape (n_peaks,)
        The m/z values of the peaks in the spectrum.
    int_array : torch.Tensor of shape (n_peaks,)
        The intensity values of the peaks in the spectrum.
    precursor_mz : float
        The precursor m/z.
    precursor_charge : int
        The precursor charge.
    tol : float, optional
        The tolerance to remove the precursor peak in m/z.

    Returns
    -------
    mz_array : torch.Tensor of shape (n_peaks,)
        The m/z values of the peaks in the spectrum.
    int_array : torch.Tensor of shape (n_peaks,)
        The intensity values of the peaks in the spectrum.
    """
    bounds = (precursor_mz - tol, precursor_mz + tol)
    outside = (mz_array >= bounds[0]) | (mz_array <= bounds[1])
    return mz_array[outside], int_array[outside]



"""Data loaders for the de novo sequencing task."""
import functools
import os
from typing import List, Optional, Tuple

import numpy as np
import pytorch_lightning as pl
import torch
from depthcharge.data import AnnotatedSpectrumIndex

from ..data.datasets import AnnotatedSpectrumDataset, SpectrumDataset


class DeNovoDataModule(pl.LightningDataModule):
    """
    Data loader to prepare MS/MS spectra for a Spec2Pep predictor.

    Parameters
    ----------
    train_index : Optional[AnnotatedSpectrumIndex]
        The spectrum index file corresponding to the training data.
    valid_index : Optional[AnnotatedSpectrumIndex]
        The spectrum index file corresponding to the validation data.
    test_index : Optional[AnnotatedSpectrumIndex]
        The spectrum index file corresponding to the testing data.
    batch_size : int
        The batch size to use for training and evaluating.
    n_peaks : Optional[int]
        The number of top-n most intense peaks to keep in each spectrum. `None`
        retains all peaks.
    min_mz : float
        The minimum m/z to include. The default is 140 m/z, in order to exclude
        TMT and iTRAQ reporter ions.
    max_mz : float
        The maximum m/z to include.
    min_intensity : float
        Remove peaks whose intensity is below `min_intensity` percentage of the
        base peak intensity.
    remove_precursor_tol : float
        Remove peaks within the given mass tolerance in Dalton around the
        precursor mass.
    n_workers : int, optional
        The number of workers to use for data loading. By default, the number of
        available CPU cores on the current machine is used.
    random_state : Optional[int]
        The NumPy random state. ``None`` leaves mass spectra in the order they
        were parsed.
    """

    def __init__(
        self,
        train_index: Optional[AnnotatedSpectrumIndex] = None,
        valid_index: Optional[AnnotatedSpectrumIndex] = None,
        test_index: Optional[AnnotatedSpectrumIndex] = None,
        batch_size: int = 128,
        n_peaks: Optional[int] = 150,
        min_mz: float = 50.0,
        max_mz: float = 2500.0,
        min_intensity: float = 0.01,
        remove_precursor_tol: float = 2.0,
        n_workers: Optional[int] = None,
        random_state: Optional[int] = None,
    ):
        super().__init__()
        self.train_index = train_index
        self.valid_index = valid_index
        self.test_index = test_index
        self.batch_size = batch_size
        self.n_peaks = n_peaks
        self.min_mz = min_mz
        self.max_mz = max_mz
        self.min_intensity = min_intensity
        self.remove_precursor_tol = remove_precursor_tol
        self.n_workers = n_workers if n_workers is not None else os.cpu_count()
        self.rng = np.random.default_rng(random_state)
        self.train_dataset = None
        self.valid_dataset = None
        self.test_dataset = None

    def setup(self, stage: str = None, annotated: bool = True) -> None:
        """
        Set up the PyTorch Datasets.

        Parameters
        ----------
        stage : str {"fit", "validate", "test"}
            The stage indicating which Datasets to prepare. All are prepared by
            default.
        annotated: bool
            True if peptide sequence annotations are available for the test
            data.
        """
        if stage in (None, "fit", "validate"):
            make_dataset = functools.partial(
                AnnotatedSpectrumDataset,
                n_peaks=self.n_peaks,
                min_mz=self.min_mz,
                max_mz=self.max_mz,
                min_intensity=self.min_intensity,
                remove_precursor_tol=self.remove_precursor_tol,
            )
            if self.train_index is not None:
                self.train_dataset = make_dataset(
                    self.train_index,
                    random_state=self.rng,
                )
            if self.valid_index is not None:
                self.valid_dataset = make_dataset(self.valid_index)
        if stage in (None, "test"):
            make_dataset = functools.partial(
                AnnotatedSpectrumDataset if annotated else SpectrumDataset,
                n_peaks=self.n_peaks,
                min_mz=self.min_mz,
                max_mz=self.max_mz,
                min_intensity=self.min_intensity,
                remove_precursor_tol=self.remove_precursor_tol,
            )
            if self.test_index is not None:
                self.test_dataset = make_dataset(self.test_index)

    def _make_loader(
        self, dataset: torch.utils.data.Dataset
    ) -> torch.utils.data.DataLoader:
        """
        Create a PyTorch DataLoader.

        Parameters
        ----------
        dataset : torch.utils.data.Dataset
            A PyTorch Dataset.

        Returns
        -------
        torch.utils.data.DataLoader
            A PyTorch DataLoader.
        """
        return torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            collate_fn=prepare_batch,
            pin_memory=True,
            num_workers=self.n_workers,
        )

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        """Get the training DataLoader."""
        return self._make_loader(self.train_dataset)

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        """Get the validation DataLoader."""
        return self._make_loader(self.valid_dataset)

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        """Get the test DataLoader."""
        return self._make_loader(self.test_dataset)

    def predict_dataloader(self) -> torch.utils.data.DataLoader:
        """Get the predict DataLoader."""
        return self._make_loader(self.test_dataset)


def prepare_batch(
    batch: List[Tuple[torch.Tensor, float, int, str]]
) -> Tuple[torch.Tensor, torch.Tensor, np.ndarray]:
    """
    Collate MS/MS spectra into a batch.

    The MS/MS spectra will be padded so that they fit nicely as a tensor.
    However, the padded elements are ignored during the subsequent steps.

    Parameters
    ----------
  batch : List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, float, int, str]]
    A batch of data from an AnnotatedSpectrumDataset, consisting of:
    (i)   A tensor with the m/z and intensity peak values, (ii)  A tensor of the precursor profile (MS1).
    (iii) A tensor of the padded product b-ion peptide, (iv)  A tensor of the padded product y-ion peptide.
    (v)   The precursor m/z value (float), (vi)  The precursor charge (int), (vii) The spectrum identifier (str).

    Returns
    -------
    spectra : torch.Tensor of shape (batch_size, n_peaks, 2)
        The padded mass spectra tensor with the m/z and intensity peak values
        for each spectrum.
    ms1: torch.Tensor of shape (batch_size, n_neighbors)
         The padded precursor elusion profile
    b_ion: torch.Tensor of shape (batch_size, peptide_len, vocab_size, n_ion_types, n_neighbors, window_size)
         The padded product b ion peptide
    y_ion: torch.Tensor of shape (batch_size, peptide_len, vocab_size, n_ion_types, n_neighbors, window_size)
        The padded product y ion peptide
    precursors : torch.Tensor of shape (batch_size, 3)
        A tensor with the precursor neutral mass, precursor charge, and
        precursor m/z.
    spectrum_ids : np.ndarray
        The spectrum identifiers (during de novo sequencing) or peptide
        sequences (during training).
    """
    spectra, ms1, b_ion, y_ion, precursor_mzs, precursor_charges, spectrum_ids = list(zip(*batch))
    spectra = torch.nn.utils.rnn.pad_sequence(spectra, batch_first=True)
    b_ion = torch.nn.utils.rnn.pad_sequence(torch.tensor(np.array(b_ion)), batch_first=True)
    y_ion = torch.nn.utils.rnn.pad_sequence(torch.tensor(np.array(y_ion)), batch_first=True)
    ms1 = torch.tensor(np.array(ms1))
    precursor_mzs = torch.tensor(precursor_mzs)
    precursor_charges = torch.tensor(precursor_charges)
    precursor_masses = (precursor_mzs - 1.007276) * precursor_charges
    precursors = torch.vstack(
        [precursor_masses, precursor_charges, precursor_mzs]
    ).T.float()
    return spectra, ms1, b_ion, y_ion, precursors, np.asarray(spectrum_ids)



"""Mass spectrometry data parsers"""
import logging
from pathlib import Path
from abc import ABC, abstractmethod

import numpy as np
from tqdm.auto import tqdm
from pyteomics.mzml import MzML
from pyteomics.mzxml import MzXML
from pyteomics.mgf import MGF


LOGGER = logging.getLogger(__name__)


class BaseParser(ABC):
    """A base parser class to inherit from.

    Parameters
    ----------
    ms_data_file : str or Path
        The mzML file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    id_type : str, optional
        The Hupo-PSI prefix for the spectrum identifier.
    """

    def __init__(
        self,
        ms_data_file,
        ms_level,
        valid_charge=None,
        id_type="scan",
    ):
        """Initialize the BaseParser"""
        self.path = Path(ms_data_file)
        self.ms_level = ms_level
        self.valid_charge = None if valid_charge is None else set(valid_charge)
        self.id_type = id_type
        self.offset = None
        self.precursor_mz = []
        self.precursor_charge = []
        self.scan_id = []
        self.mz_arrays = []
        self.intensity_arrays = []
        self.scans = []

    @abstractmethod
    def open(self):
        """Open the file as an iterable"""
        pass

    @abstractmethod
    def parse_spectrum(self, spectrum):
        """Parse a single spectrum

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in a given format.
        """
        pass

    def read(self):
        """Read the ms data file"""
        n_skipped = 0
        with self.open() as spectra:
            for spectrum in tqdm(spectra, desc=str(self.path), unit="spectra"):
                try:
                    self.parse_spectrum(spectrum)
                except (IndexError, KeyError, ValueError):
                    n_skipped += 1

        if n_skipped:
            LOGGER.warning(
                "Skipped %d spectra with invalid precursor info", n_skipped
            )

        self.precursor_mz = np.array(self.precursor_mz, dtype=np.float64)
        self.precursor_charge = np.array(
            self.precursor_charge,
            dtype=np.uint8,
        )

        self.scan_id = np.array(self.scan_id)

        # Build the index
        sizes = np.array([0] + [s.shape[0] for s in self.mz_arrays])
        self.offset = sizes[:-1].cumsum()
        self.mz_arrays = np.concatenate(self.mz_arrays).astype(np.float64)
        self.intensity_arrays = np.concatenate(self.intensity_arrays).astype(
            np.float32
        )

    @property
    def n_spectra(self):
        """The number of spectra"""
        return self.offset.shape[0]

    @property
    def n_peaks(self):
        """The number of peaks in the file."""
        return self.mz_arrays.shape[0]


class MzmlParser(BaseParser):
    """Parse mass spectra from an mzML file.

    Parameters
    ----------
    ms_data_file : str or Path
        The mzML file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    """

    def __init__(self, ms_data_file, ms_level=2, valid_charge=None):
        """Initialize the MzmlParser."""
        super().__init__(
            ms_data_file,
            ms_level=ms_level,
            valid_charge=valid_charge,
        )

    def open(self):
        """Open the mzML file for reading"""
        return MzML(str(self.path))

    def parse_spectrum(self, spectrum):
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in mzML format.
        """
        if spectrum["ms level"] != self.ms_level:
            return

        if self.ms_level > 1:
            precursor = spectrum["precursorList"]["precursor"][0]
            precursor_ion = precursor["selectedIonList"]["selectedIon"][0]
            precursor_mz = float(precursor_ion["selected ion m/z"])
            if "charge state" in precursor_ion:
                precursor_charge = int(precursor_ion["charge state"])
            elif "possible charge state" in precursor_ion:
                precursor_charge = int(precursor_ion["possible charge state"])
            else:
                precursor_charge = 0
        else:
            precursor_mz, precursor_charge = None, 0

        if self.valid_charge is None or precursor_charge in self.valid_charge:
            self.mz_arrays.append(spectrum["m/z array"])
            self.intensity_arrays.append(spectrum["intensity array"])
            self.precursor_mz.append(precursor_mz)
            self.precursor_charge.append(precursor_charge)
            self.scan_id.append(_parse_scan_id(spectrum["id"]))


class MzxmlParser(BaseParser):
    """Parse mass spectra from an mzXML file.

    Parameters
    ----------
    ms_data_file : str or Path
        The mzXML file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    """

    def __init__(self, ms_data_file, ms_level=2, valid_charge=None):
        """Initialize the MzxmlParser."""
        super().__init__(
            ms_data_file,
            ms_level=ms_level,
            valid_charge=valid_charge,
        )

    def open(self):
        """Open the mzXML file for reading"""
        return MzXML(str(self.path))

    def parse_spectrum(self, spectrum):
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in mzXML format.
        """
        if spectrum["msLevel"] != self.ms_level:
            return

        if self.ms_level > 1:
            precursor = spectrum["precursorMz"][0]
            precursor_mz = float(precursor["precursorMz"])
            precursor_charge = int(precursor.get("precursorCharge", 0))
        else:
            precursor_mz, precursor_charge = None, 0

        if self.valid_charge is None or precursor_charge in self.valid_charge:
            self.mz_arrays.append(spectrum["m/z array"])
            self.intensity_arrays.append(spectrum["intensity array"])
            self.precursor_mz.append(precursor_mz)
            self.precursor_charge.append(precursor_charge)
            self.scan_id.append(_parse_scan_id(spectrum["id"]))


class MgfParser(BaseParser):
    """Parse mass spectra from an MGF file.

    Parameters
    ----------
    ms_data_file : str or Path
        The MGF file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    annotations : bool
        Include peptide annotations.
    """

    def __init__(
        self,
        ms_data_file,
        ms_level=2,
        valid_charge=None,
        annotations=False,
    ):
        """Initialize the MgfParser."""
        super().__init__(
            ms_data_file,
            ms_level=ms_level,
            valid_charge=valid_charge,
            id_type="index",
        )
        self.annotations = [] if annotations else None
        self._counter = 0

    def open(self):
        """Open the MGF file for reading"""
        return MGF(str(self.path))

    def parse_spectrum(self, spectrum):
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in MGF format.
        """
        if self.ms_level > 1:
            precursor_mz = float(spectrum["params"]["pepmass"][0])
            precursor_charge = int(spectrum["params"].get("charge", [0])[0])
        else:
            precursor_mz, precursor_charge = None, 0

        if self.annotations is not None:
            self.annotations.append(spectrum["params"].get("seq"))

        if self.valid_charge is None or precursor_charge in self.valid_charge:
            self.mz_arrays.append(spectrum["m/z array"])
            self.scans.append(spectrum["params"]["scans"])
            self.intensity_arrays.append(spectrum["intensity array"])
            self.precursor_mz.append(precursor_mz)
            self.precursor_charge.append(precursor_charge)
            self.scan_id.append(self._counter)

        self._counter += 1


def _parse_scan_id(scan_str):
    """Remove the string prefix from the scan ID.

    Adapted from:
    https://github.com/bittremieux/GLEAMS/blob/
    8831ad6b7a5fc391f8d3b79dec976b51a2279306/gleams/
    ms_io/mzml_io.py#L82-L85

    Parameters
    ----------
    scan_str : str
        The scan ID string.

    Returns
    -------
    int
        The scan ID number.
    """
    try:
        return int(scan_str)
    except ValueError:
        try:
            return int(scan_str[scan_str.find("scan=") + len("scan=") :])
        except ValueError:
            pass

    raise ValueError(f"Failed to parse scan number")



"""Amino acid masses and other useful mass spectrometry calculations"""
import re


class PeptideMass:
    """A simple class for calculating peptide masses

    Parameters
    ----------
    residues: Dict or str {"massivekb", "canonical"}, optional
        The amino acid dictionary and their masses. By default this is only
        the 20 canonical amino acids, with cysteine carbamidomethylated. If
        "massivekb", this dictionary will include the modifications found in
        MassIVE-KB. Additionally, a dictionary can be used to specify a custom
        collection of amino acids and masses.
    """

    canonical = {
        "G": 57.021463735,
        "A": 71.037113805,
        "S": 87.032028435,
        "P": 97.052763875,
        "V": 99.068413945,
        "T": 101.047678505,
        "C+57.021": 103.009184505 + 57.02146,
        "L": 113.084064015,
        "I": 113.084064015,
        "N": 114.042927470,
        "D": 115.026943065,
        "Q": 128.058577540,
        "K": 128.094963050,
        "E": 129.042593135,
        "M": 131.040484645,
        "H": 137.058911875,
        "F": 147.068413945,
        # "U": 150.953633405,
        "R": 156.101111050,
        "Y": 163.063328575,
        "W": 186.079312980,
        # "O": 237.147726925,
    }

    # Modfications found in MassIVE-KB
    massivekb = {
        # N-terminal mods:
        "+42.011": 42.010565,  # Acetylation
        "+43.006": 43.005814,  # Carbamylation
        "-17.027": -17.026549,  # NH3 loss
        "+43.006-17.027": (43.006814 - 17.026549),
        # AA mods:
        "M+15.995": canonical["M"] + 15.994915,  # Met Oxidation
        "N+0.984": canonical["N"] + 0.984016,  # Asn Deamidation
        "Q+0.984": canonical["Q"] + 0.984016,  # Gln Deamidation
    }

    # Constants
    hydrogen = 1.007825035
    oxygen = 15.99491463
    h2o = 2 * hydrogen + oxygen
    proton = 1.00727646688

    def __init__(self, residues="canonical"):
        """Initialize the PeptideMass object"""
        if residues == "canonical":
            self.masses = self.canonical
        elif residues == "massivekb":
            self.masses = self.canonical
            self.masses.update(self.massivekb)
        else:
            self.masses = residues

    def __len__(self):
        """Return the length of the residue dictionary"""
        return len(self.masses)

    def mass(self, seq, charge=None):
        """Calculate a peptide's mass or m/z.

        Parameters
        ----------
        seq : list or str
            The peptide sequence, using tokens defined in ``self.residues``.
        charge : int, optional
            The charge used to compute m/z. Otherwise the neutral peptide mass
            is calculated

        Returns
        -------
        float
            The computed mass or m/z.
        """
        #print(seq)
        if isinstance(seq, str):
            seq = re.split(r"(?<=.)(?=[A-Z])", seq)

        calc_mass = sum([self.masses[aa] for aa in seq if aa is not '']) + self.h2o
        if charge is not None:
            calc_mass = (calc_mass / charge) + self.proton

        return calc_mass



"""Base Transformer models for working with mass spectra and peptides"""
import re

import torch
import torch.nn as nn
import numpy as np
from .encoders import MassEncoder, PeakEncoder, PositionalEncoder
from ..masses import PeptideMass
from .. import utils
import torch.nn.functional as F
from transfusion.model  import timestep_embedding


class WindowedSelfAttention(nn.Module):
    def __init__(self, input_dim, window_size):
        super(WindowedSelfAttention, self).__init__()
        self.input_dim = input_dim
        self.window_size = window_size

    def forward(self, query, key, value):
        batch_size, query_seq_len, _ = query.size()
        _, key_seq_len, _ = key.size()

        # Initialize an empty tensor for the output
        output = torch.zeros(batch_size, query_seq_len, self.input_dim).to(query.device)

        # Calculate attention scores for each query position
        for i in range(query_seq_len):
            start = max(0, i - self.window_size // 2)
            end = min(key_seq_len, i + self.window_size // 2 + 1)
            attention_scores = torch.matmul(query[:, i, :].unsqueeze(1), key[:, start:end, :].transpose(1, 2))
            attention_scores = attention_scores / (key[:, start:end, :].size(-1) ** 0.5)
            attention_weights = F.softmax(attention_scores, dim=-1)
            weighted_sum = torch.matmul(attention_weights, value[:, start:end, :])
            output[:, i, :] = weighted_sum.squeeze(1)

        return output



class SelfAttentionConcatenation(nn.Module):
    def __init__(self, input_dim):
        super(SelfAttentionConcatenation, self).__init__()

        self.attention = nn.MultiheadAttention(input_dim, num_heads=1)

    def forward(self, *tensors):
        attn_tensors = []

        for tensor in tensors:
            attn_tensor, _ = self.attention(tensor, tensor, tensor)
            attn_tensors.append(attn_tensor)

        # Transpose and reshape the attention tensors to match the desired output shape
        output_dim = attn_tensors[0].shape[-1]
        attn_tensors = [attn_tensor.transpose(0, 1).reshape(-1, attn_tensor.shape[-2], output_dim) for attn_tensor in
                        attn_tensors]

        # Concatenate the tensors along the second dimension
        combined_tensor = torch.cat(attn_tensors, dim=1)

        return combined_tensor

class AttentionConcatenation(nn.Module):
    def __init__(self, input_dim):
        super(AttentionConcatenation, self).__init__()

        self.attention = nn.Sequential(
            nn.Linear(input_dim, 1),
            nn.Tanh(),
            nn.Softmax(dim=1)  # Apply softmax along the second dimension
        )

    def forward(self, tensor1, tensor2, tensor3):
        attn_weights1 = self.attention(tensor1)
        attn_weights2 = self.attention(tensor2)
        attn_weights3 = self.attention(tensor3)

        weighted_tensor1 = torch.mul(tensor1, attn_weights1)
        weighted_tensor2 = torch.mul(tensor2, attn_weights2)
        weighted_tensor3 = torch.mul(tensor3, attn_weights3)

        concatenated_tensor = torch.cat([weighted_tensor1, weighted_tensor2, weighted_tensor3], dim=1)

        return  concatenated_tensor #weighted_tensor1 #concatenated_tensor


class SpectrumEncoder(torch.nn.Module):
    """A Transformer encoder for input mass spectra.

    Parameters
    ----------
    dim_model : int, optional
        The latent dimensionality to represent peaks in the mass spectrum.
    n_head : int, optional
        The number of attention heads in each layer. ``dim_model`` must be
        divisible by ``n_head``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        layers of the model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    peak_encoder : bool, optional
        Use positional encodings m/z values of each peak.
    dim_intensity: int or None, optional
        The number of features to use for encoding peak intensity.
        The remaining (``dim_model - dim_intensity``) are reserved for
        encoding the m/z value.
    """

    def __init__(
            self,
            dim_model=128,
            vocab_size=26,
            win_size=10,
            num_ions=8,
            num_neighbor=5,
            n_head=8,
            dim_feedforward=1024,
            n_layers=1,
            dropout=0,
            peak_encoder=True,
            dim_intensity=None,
    ):
        """Initialize a SpectrumEncoder"""
        super().__init__()

        self.latent_spectrum = torch.nn.Parameter(torch.randn(1, 1, dim_model))
        window_size = 5  # the ms1 profile shape
        # self.attention_layer = AttentionConcatenation(dim_model) #SelfAttentionConcatenation(dim_model)  #WindowedSelfAttention(dim_model, window_size)
        # SelfAttentionConcatenation(dim_model) #AttentionConcatenation(dim_model)

        if peak_encoder:
            self.peak_encoder = PeakEncoder(
                dim_model,
                dim_intensity=dim_intensity,
            )
        else:
            self.peak_encoder = torch.nn.Linear(2, dim_model)

        self.ms1_encoder = torch.nn.Linear(1, dim_model)
        # multiple by two because of b ion and y ion
        self.ms2_encoder = torch.nn.Linear(vocab_size * win_size * num_ions * num_neighbor, dim_model)

        # The Transformer layers:
        layer = torch.nn.TransformerEncoderLayer(
            d_model=dim_model,
            nhead=n_head,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            dropout=dropout,
        )

        self.transformer_encoder = torch.nn.TransformerEncoder(
            layer,
            num_layers=n_layers,
        )

    def mask1(self, X):
        zeros = ~X.sum(dim=2).bool()

        return zeros

    def calculate_mask(self, X):

        zeros = ~X.sum(dim=2).bool()
        mask = [
            torch.tensor([[False]] * X.shape[0]).type_as(zeros),
            zeros,
        ]

        mask = torch.cat(mask, dim=1)

        return mask

    def forward(self, spectra, ms1, ms2_b_ion, ms2_y_ion):
        """The forward pass.

        Parameters
        ----------
        spectra : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.
         ms1: torch.Tensor of shape (n_spectra, n_neghibors)
         ms2_b_ion: torch.Tensor of shape (n_spectra, seq_len, vocab_size, n_ion, n_neighbors, window_size)
         ms2_y_ion: torch.Tensor of shape (n_spectra, seq_len, vocab_size, n_ions, n_neighbors, window_size)
        Returns
        -------
        latent : torch.Tensor of shape (n_spectra, n_peaks + 1, dim_model)
            The latent representations for the spectrum and each of its
            peaks.
        mem_mask : torch.Tensor
            The memory mask specifying which elements were padding in X.
        """
        zeros = ~spectra.sum(dim=2).bool()
        mask = [
            torch.tensor([[False]] * spectra.shape[0]).type_as(zeros),
            zeros,
        ]
        spectra_mask = torch.cat(mask, dim=1)
        peaks = self.peak_encoder(spectra)

        # Add the spectrum representation to each input:
        latent_spectra = self.latent_spectrum.expand(peaks.shape[0], -1, -1)

        ms1 = ms1[:, :, None]
        ms1_mask = self.calculate_mask(ms1)
        ms1 = self.ms1_encoder(ms1.to(self.ms1_encoder.weight.dtype))

        ms2_b_ion = ms2_b_ion.view(ms2_b_ion.size(0), ms2_b_ion.size(1), -1)
        ms2_b_ion_mask = self.calculate_mask(ms2_b_ion)
        ms2_b_ion = self.ms2_encoder(ms2_b_ion.to(self.ms2_encoder.weight.dtype))

        ms2_y_ion = ms2_y_ion.view(ms2_y_ion.size(0), ms2_y_ion.size(1), -1)
        ms2_y_ion_mask = self.calculate_mask(ms2_y_ion)
        ms2_y_ion = self.ms2_encoder(ms2_y_ion.to(self.ms2_encoder.weight.dtype))

        ms2_b_ion = torch.cat([latent_spectra, ms2_b_ion], dim=1)
        ms2_y_ion = torch.cat([latent_spectra, ms2_y_ion], dim=1)
        ms2 = torch.cat([ms2_b_ion, ms2_y_ion], dim=1)

        # ms2 = torch.cat([ms2_b_ion, ms2_y_ion], dim =1)
        # ms2 = self.ms2_encoder(ms2.to(self.ms2_encoder.weight.dtype))

        ms2_mask = torch.cat([ms2_b_ion_mask, ms2_y_ion_mask], dim=1)

        ms1 = torch.cat([latent_spectra, ms1], dim=1)
        # ms2 = torch.cat([latent_spectra, ms2], dim=1)
        mask = torch.cat([spectra_mask, ms2_mask, ms1_mask], dim=1)

        peaks = torch.cat([latent_spectra, peaks], dim=1)
        peaks = torch.cat([peaks, ms2, ms1], dim=1)
        # peaks = self.windowed_attention(peaks, ms1, ms2)
        # peaks = self.attention_layer(peaks, ms2, ms1)

        return self.transformer_encoder(peaks, src_key_padding_mask=mask), mask

    @property
    def device(self):
        """The current device for the model"""
        return next(self.parameters()).device


class _PeptideTransformer(torch.nn.Module):
    """A transformer base class for peptide sequences.

    Parameters
    ----------
    dim_model : int
        The latent dimensionality to represent the amino acids in a peptide
        sequence.
    pos_encoder : bool
        Use positional encodings for the amino acid sequence.
    residues: Dict or str {"massivekb", "canonical"}, optional
        The amino acid dictionary and their masses. By default this is only
        the 20 canonical amino acids, with cysteine carbamidomethylated. If
        "massivekb", this dictionary will include the modifications found in
        MassIVE-KB. Additionally, a dictionary can be used to specify a custom
        collection of amino acids and masses.
    max_charge : int
        The maximum charge to embed.
    """

    def __init__(
        self,
        dim_model,
        pos_encoder,
        residues,
        max_charge,
    ):
        super().__init__()
        self.reverse = False
        self._peptide_mass = PeptideMass(residues=residues)
        self._amino_acids = list(self._peptide_mass.masses.keys()) + ["$"]
        self._idx2aa = {i + 1: aa for i, aa in enumerate(self._amino_acids)}
        self._aa2idx = {aa: i for i, aa in self._idx2aa.items()}

        if pos_encoder:
            self.pos_encoder = PositionalEncoder(dim_model)
        else:
            self.pos_encoder = torch.nn.Identity()

        self.charge_encoder = torch.nn.Embedding(max_charge, dim_model)
        self.aa_encoder = torch.nn.Embedding(
            len(self._amino_acids) + 1,
            dim_model,
            padding_idx=0,
        )

    def tokenize(self, sequence, partial=False):
        """Transform a peptide sequence into tokens

        Parameters
        ----------
        sequence : str
            A peptide sequence.

        Returns
        -------
        torch.Tensor
            The token for each amino acid in the peptide sequence.
        """
        if not isinstance(sequence, str):
            return sequence  # Assume it is already tokenized.

        sequence = sequence.replace("I", "L")
        sequence = re.split(r"(?<=.)(?=[A-Z])", sequence)
        if self.reverse:
            sequence = list(reversed(sequence))

        if not partial:
            sequence += ["$"]

        tokens = [self._aa2idx[aa] for aa in sequence]
        tokens = torch.tensor(tokens, device=self.device)
        return tokens

    def detokenize(self, tokens):
        """Transform tokens back into a peptide sequence.

        Parameters
        ----------
        tokens : torch.Tensor of shape (n_amino_acids,)
            The token for each amino acid in the peptide sequence.

        Returns
        -------
        list of str
            The amino acids in the peptide sequence.
        """
        sequence = [self._idx2aa.get(i.item(), "") for i in tokens]
        if "$" in sequence:
            idx = sequence.index("$")
            sequence = sequence[: idx + 1]

        if self.reverse:
            sequence = list(reversed(sequence))

        return sequence

    @property
    def vocab_size(self):
        """Return the number of amino acids"""
        return len(self._aa2idx)

    @property
    def device(self):
        """The current device for the model"""
        return next(self.parameters()).device


class PeptideEncoder(_PeptideTransformer):
    """A transformer encoder for peptide sequences.

    Parameters
    ----------
    dim_model : int
        The latent dimensionality to represent the amino acids in a peptide
        sequence.
    n_head : int, optional
        The number of attention heads in each layer. ``dim_model`` must be
        divisible by ``n_head``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        layers of the model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    pos_encoder : bool, optional
        Use positional encodings for the amino acid sequence.
    residues: Dict or str {"massivekb", "canonical"}, optional
        The amino acid dictionary and their masses. By default this is only
        the 20 canonical amino acids, with cysteine carbamidomethylated. If
        "massivekb", this dictionary will include the modifications found in
        MassIVE-KB. Additionally, a dictionary can be used to specify a custom
        collection of amino acids and masses.
    max_charge : int, optional
        The maximum charge state for peptide sequences.
    """

    def __init__(
        self,
        dim_model=128,
        n_head=8,
        dim_feedforward=1024,
        n_layers=1,
        dropout=0,
        pos_encoder=True,
        residues="canonical",
        max_charge=5,
    ):
        """Initialize a PeptideEncoder"""
        super().__init__(
            dim_model=dim_model,
            pos_encoder=pos_encoder,
            residues=residues,
            max_charge=max_charge,
        )

        # The Transformer layers:
        layer = torch.nn.TransformerEncoderLayer(
            d_model=dim_model,
            nhead=n_head,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            dropout=dropout,
        )

        self.transformer_encoder = torch.nn.TransformerEncoder(
            layer,
            num_layers=n_layers,
        )

    def forward(self, sequences, charges):
        """Predict the next amino acid for a collection of sequences.

        Parameters
        ----------
        sequences : list of str or list of torch.Tensor of length batch_size
            The partial peptide sequences for which to predict the next
            amino acid. Optionally, these may be the token indices instead
            of a string.
        charges : torch.Tensor of size (batch_size,)
            The charge state of the peptide

        Returns
        -------
        latent : torch.Tensor of shape (n_sequences, len_sequence, dim_model)
            The latent representations for the spectrum and each of its
            peaks.
        mem_mask : torch.Tensor
            The memory mask specifying which elements were padding in X.
        """
        sequences = utils.listify(sequences)
        tokens = [self.tokenize(s) for s in sequences]
        tokens = torch.nn.utils.rnn.pad_sequence(tokens, batch_first=True)
        encoded = self.aa_encoder(tokens)

        # Encode charges
        charges = self.charge_encoder(charges - 1)[:, None]
        encoded = torch.cat([charges, encoded], dim=1)

        # Create mask
        mask = ~encoded.sum(dim=2).bool()

        # Add positional encodings
        encoded = self.pos_encoder(encoded)

        # Run through the model:
        latent = self.transformer_encoder(encoded, src_key_padding_mask=mask)
        return latent, mask


class PeptideDecoder(_PeptideTransformer):
    """A transformer decoder for peptide sequences.

    Parameters
    ----------
    dim_model : int, optional
        The latent dimensionality to represent peaks in the mass spectrum.
    n_head : int, optional
        The number of attention heads in each layer. ``dim_model`` must be
        divisible by ``n_head``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        layers of the model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    pos_encoder : bool, optional
        Use positional encodings for the amino acid sequence.
    reverse : bool, optional
        Sequence peptides from c-terminus to n-terminus.
    residues: Dict or str {"massivekb", "canonical"}, optional
        The amino acid dictionary and their masses. By default this is only
        the 20 canonical amino acids, with cysteine carbamidomethylated. If
        "massivekb", this dictionary will include the modifications found in
        MassIVE-KB. Additionally, a dictionary can be used to specify a custom
        collection of amino acids and masses.
    """

    def __init__(
        self,
        dim_model=128,
        n_head=8,
        dim_feedforward=1024,
        n_layers=1,
        dropout=0,
        pos_encoder=True,
        reverse=True,
        residues="canonical",
        max_charge=5,
        t_emb_dim = 128,
        t_emb_max_period: int = 10000
    ):
        """Initialize a PeptideDecoder"""
        super().__init__(
            dim_model=dim_model,
            pos_encoder=pos_encoder,
            residues=residues,
            max_charge=max_charge,
        )
        self.reverse = reverse
        self.t_emb_max_period = t_emb_max_period
        self.t_emb_dim = t_emb_dim

        # Additional model components
        self.mass_encoder = MassEncoder(dim_model)
        layer = torch.nn.TransformerDecoderLayer(
            d_model=dim_model,
            nhead=n_head,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            dropout=dropout,
        )
         
        self.transformer_decoder = torch.nn.TransformerDecoder(
            layer,
            num_layers=n_layers,
        )

        self.t_emb_max_period = t_emb_max_period
        
        self.t_emb_dim = t_emb_dim
        
        self.t_layers = nn.Sequential(
            nn.Linear(t_emb_dim, dim_model),
            nn.SiLU(),
            nn.Linear(dim_model, dim_model)
        )
        self.final = torch.nn.Linear(dim_model, len(self._amino_acids) + 1)

    def forward(self, sequences, time_, precursors, memory, memory_key_padding_mask):
        """Predict the next amino acid for a collection of sequences.

        Parameters
        ----------
        sequences : list of str or list of torch.Tensor
            The partial peptide sequences for which to predict the next
            amino acid. Optionally, these may be the token indices instead
            of a string.
        time_: long tesnor of timestep indices
        precursors : torch.Tensor of size (batch_size, 2)
            The measured precursor mass (axis 0) and charge (axis 1) of each
            tandem mass spectrum
        memory : torch.Tensor of shape (batch_size, n_peaks, dim_model)
            The representations from a ``TransformerEncoder``, such as a
           ``SpectrumEncoder``.
        memory_key_padding_mask : torch.Tensor of shape (batch_size, n_peaks)
            The mask that indicates which elements of ``memory`` are padding.

        Returns
        -------
        scores : torch.Tensor of size (batch_size, len_sequence, n_amino_acids)
            The raw output for the final linear layer. These can be Softmax
            transformed to yield the probability of each amino acid for the
            prediction.
        tokens : torch.Tensor of size (batch_size, len_sequence)
            The input padded tokens.

        """
        # Prepare sequences
        if sequences is not None:
            sequences = utils.listify(sequences)
            tokens = [self.tokenize(s) for s in sequences]
            tokens = torch.nn.utils.rnn.pad_sequence(tokens, batch_first=True)
        else:
            tokens = torch.tensor([[]]).to(self.device)

        t = 0 
        if time_ is not None:
            self.t_emb = timestep_embedding(
                time_, self.t_emb_dim, self.t_emb_max_period, dtype=memory.dtype
            )
            t = self.t_layers(self.t_emb)[:, None]  # (bs, 1, dim)
        # Prepare mass and charge
        masses = self.mass_encoder(precursors[:, None, [0]])
        charges = self.charge_encoder(precursors[:, 1].int() - 1)
        precursors = masses + charges[:, None, :]

        # Feed through model:
        if sequences is None:
            tgt = precursors
        else:
            tgt = torch.cat([precursors, self.aa_encoder(tokens)], dim=1)

        tgt_key_padding_mask = tgt.sum(axis=2) == 0
        tgt = self.pos_encoder(tgt)
        tgt_mask = generate_tgt_mask(tgt.shape[1]).type_as(precursors)
        tgt = t + tgt
        preds = self.transformer_decoder(
            tgt=tgt,
            memory=memory,
            tgt_mask=tgt_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=memory_key_padding_mask.to(self.device),
        )
        return self.final(preds), tokens


def generate_tgt_mask(sz):
    """Generate a square mask for the sequence. The masked positions
    are filled with float('-inf'). Unmasked positions are filled with
    float(0.0).

    This function is a slight modification of the version in the PyTorch
    repository.

    Parameters
    ----------
    sz : int
        The length of the target sequence.
    """
    mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
    mask = (
        mask.float()
        .masked_fill(mask == 0, float("-inf"))
        .masked_fill(mask == 1, float(0.0))
    )
    return mask



"""A PyTorch Dataset class for annotated spectra."""
import torch
import numpy as np
from torch.utils.data import Dataset

from . import preprocessing
from .. import utils


class SpectrumDataset(Dataset):
    """Parse and retrieve collections of mass spectra.

    Parameters
    ----------
    spectrum_index : depthcharge.data.SpectrumIndex
        The collection of spectra to use as a dataset.
    n_peaks : int, optional
        Keep only the top-n most intense peaks in any spectrum. ``None``
        retains all of the peaks.
    min_mz : float, optional
        The minimum m/z to include. The default is 140 m/z, in order to
        exclude TMT and iTRAQ reporter ions.
    preprocessing_fn: Callable or list of Callable, optional
        The function(s) used to preprocess the mass spectra. See the
        preprocessing module for details. ``None``, the default, square root
        transforms the intensities and scales them to unit norm.

    Attributes
    ----------
    n_peaks : int
        The maximum number of mass speak to consider for each mass spectrum.
    min_mz : float
        The minimum m/z to consider for each mass spectrum.
    n_spectra : int
    index : depthcharge.data.SpectrumIndex
    """

    def __init__(
        self,
        spectrum_index,
        n_peaks=200,
        min_mz=140,
        preprocessing_fn=None,
    ):
        """Initialize a SpectrumDataset"""
        super().__init__()
        self.n_peaks = n_peaks
        self.min_mz = min_mz
        self.preprocessing_fn = preprocessing_fn
        self._index = spectrum_index

    def __len__(self):
        """The number of spectra."""
        return self.n_spectra

    def __getitem__(self, idx):
        """Return a single mass spectrum.

        Parameters
        ----------
        idx : int
            The index to return.

        Returns
        -------
        spectrum : torch.Tensor of shape (n_peaks, 2)
            The mass spectrum where ``spectrum[:, 0]`` are the m/z values and
            ``spectrum[:, 1]`` are their associated intensities.
        precursor_mz : float
            The m/z of the precursor.
        precursor_charge : int
            The charge of the precursor.
        """
        batch = self.index[idx]
        spec = self._process_peaks(*batch)
        if not spec.sum():
            spec = torch.tensor([[0, 1]]).float()

        return spec, batch[2], batch[3]

    def get_spectrum_id(self, idx):
        """Return the identifier for a single mass spectrum.

        Parameters
        ----------
        idx : int
            The index of a mass spectrum within the SpectrumIndex.

        Returns
        -------
        ms_data_file : str
            The mass spectrometry data file from which the mass spectrum was
            originally parsed.
        identifier : str
            The mass spectrum identifier, per PSI recommendations.
        """
        with self.index:
            return self.index.get_spectrum_id(idx)

    def _process_peaks(
        self,
        mz_array,
        int_array,
        precursor_mz,
        precursor_charge,
    ):
        """Process each mass spectrum.

        Parameters
        ----------
        mz_array : numpy.ndarray of shape (n_peaks,)
            The m/z values of the peaks in the spectrum.
        int_array : numpy.ndarray of shape (n_peaks,)
            The intensity values of the peaks in the spectrum.
        precursor_mz : float
            The precursor m/z.
        precursor_charge : int
            The precursor charge.

        Returns
        -------
        torch.Tensor of shape (n_peaks, 2)
            The mass spectrum where column 0 are the m/z values and
            column 1 are their associated intensities.
        """
        if self.min_mz is not None:
            keep = mz_array >= self.min_mz
            mz_array = mz_array[keep]
            int_array = int_array[keep]

        if len(int_array) > self.n_peaks:
            top_p = np.argpartition(int_array, -self.n_peaks)[-self.n_peaks :]
            top_p = np.sort(top_p)
            mz_array = mz_array[top_p]
            int_array = int_array[top_p]

        mz_array = torch.tensor(mz_array)
        int_array = torch.tensor(int_array)
        for func in self.preprocessing_fn:
            mz_array, int_array = func(
                mz_array=mz_array,
                int_array=int_array,
                precursor_mz=precursor_mz,
                precursor_charge=precursor_charge,
            )

        return torch.vstack([mz_array, int_array]).T.float()

    @property
    def n_spectra(self):
        """The total number of spectra."""
        return self.index.n_spectra

    @property
    def index(self):
        """The underyling SpectrumIndex."""
        return self._index

    @property
    def preprocessing_fn(self):
        """The functions to preprocess each mass spectrum."""
        return self._preprocessing_fn

    @preprocessing_fn.setter
    def preprocessing_fn(self, funcs):
        """Set the preprocessing_fn"""
        if funcs is None:
            funcs = preprocessing.sqrt_and_norm

        self._preprocessing_fn = utils.listify(funcs)

    @staticmethod
    def collate_fn(batch):
        """This is the collate function for a SpectrumDataset.

        The mass spectra must be padded so that they fit nicely as a tensor.
        However, the padded elements are ignored during the subsequent steps.

        Parameters
        ----------
        batch : tuple of tuple of torch.Tensor
            A batch of data from an AnnotatedSpectrumDataset.

        Returns
        -------
        spectra : torch.Tensor of shape (batch_size, n_peaks, 2)
            The mass spectra to sequence, where ``X[:, :, 0]`` are the m/z
            values and ``X[:, :, 1]`` are their associated intensities.
        precursors : torch.Tensor of shape (batch_size, 2)
            The precursor mass and charge state.
        """
        spec, mz, charge = list(zip(*batch))
        charge = torch.tensor(charge)
        mass = (torch.tensor(mz) - 1.007276) * charge
        precursors = torch.vstack([mass, charge]).T.float()
        spec = torch.nn.utils.rnn.pad_sequence(
            spec,
            batch_first=True,
        )
        return spec, precursors


class AnnotatedSpectrumDataset(SpectrumDataset):
    """Parse and retrieve collections of mass spectra

    Parameters
    ----------
    annotated_spectrum_index : depthcharge.data.SpectrumIndex
        The collection of annotated mass spectra to use as a dataset.
    n_peaks : int, optional
        Keep only the top-n most intense peaks in any spectrum. ``None``
        retains all of the peaks.
    min_mz : float, optional
        The minimum m/z to include. The default is 140 m/z, in order to
        exclude TMT and iTRAQ reporter ions.
    preprocessing_fn: Callable or list of Callable, optional
        The function(s) used to preprocess the mass spectra. See the
        preprocessing module for details. ``None``, the default, square root
        transforms the intensities and scales them to unit norm.

    Attributes
    ----------
    n_peaks : int
        The maximum number of mass speak to consider for each mass spectrum.
    min_mz : float
        The minimum m/z to consider for each mass spectrum.
    n_spectra : int
    index : depthcharge.data.AnnotatedSpectrumIndex
    """

    def __init__(
        self,
        annotated_spectrum_index,
        n_peaks=200,
        min_mz=140,
        preprocessing_fn=None,
    ):
        """Initialize an AnnotatedSpectrumDataset"""
        super().__init__(
            annotated_spectrum_index,
            n_peaks=n_peaks,
            min_mz=min_mz,
            preprocessing_fn=preprocessing_fn,
        )

    def __getitem__(self, idx):
        """Return a single annotated mass spectrum.

        Parameters
        ----------
        idx : int
            The index to return.

        Returns
        -------
        spectrum : torch.Tensor of shape (n_peaks, 2)
            The mass spectrum where ``spectrum[:, 0]`` are the m/z values and
            ``spectrum[:, 1]`` are their associated intensities.
        precursor_mz : float
            The m/z of the precursor.
        precursor_charge : int
            The charge of the precursor.
        annotation : str
            The annotation for the mass spectrum.
        """
        *batch, seq = self.index[idx]
        spec = self._process_peaks(*batch)
        if not spec.sum():
            spec = torch.tensor([[0, 1]]).float()

        return spec, batch[2], batch[3], seq

    @staticmethod
    def collate_fn(batch):
        """This is the collate function for an AnnotatedSpectrumDataset.

        The mass spectra must be padded so that they fit nicely as a tensor.
        However, the padded elements are ignored during the subsequent steps.

        Parameters
        ----------
        batch : tuple of tuple of torch.Tensor
            A batch of data from an AnnotatedSpectrumDataset.

        Returns
        -------
        spectra : torch.Tensor of shape (batch_size, n_peaks, 2)
            The mass spectra to sequence, where ``X[:, :, 0]`` are the m/z
            values and ``X[:, :, 1]`` are their associated intensities.
        precursors : torch.Tensor of shape (batch_size, 2)
            The precursor mass and charge state.
        sequence : list of str
            The peptide sequence annotations.

        """
        spec, mz, charge, seq = list(zip(*batch))
        charge = torch.tensor(charge)
        mass = (torch.tensor(mz) - 1.007276) * charge
        precursors = torch.vstack([mass, charge]).T.float()
        spec = torch.nn.utils.rnn.pad_sequence(spec, batch_first=True)
        return spec, precursors, np.array(seq)



"""Get the package version"""


def _get_version():
    """Retreive the package version.

    Using setuptools-scm dynamically sets the pacakge version depending on
    the latest git release and commits since that point. This function
    returns the current version number.

    Returns
    -------
    str or None
        The package version number. If not version is found, returns None.
    """
    try:
        # Fast, but only works in Python 3.8+
        from importlib.metadata import version, PackageNotFoundError

        try:
            return version("ganbit")
        except PackageNotFoundError:
            return None

    except ImportError:
        # Slow, but works for all Python 3+
        from pkg_resources import get_distribution, DistributionNotFound

        try:
            return get_distribution("ganbit").version
        except DistributionNotFound:
            return None



"""Small utility functions"""
import os
import platform
import re
from typing import Tuple

import psutil
import torch
import pickle

def dict_to_pkl(data):
    with open('true-pred-pep.pkl', 'ab+') as handle:
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)

def n_workers() -> int:
    """
    Get the number of workers to use for data loading.

    This is the maximum number of CPUs allowed for the process, scaled for the
    number of GPUs being used.

    On Windows and MacOS, we only use the main process. See:
    https://discuss.pytorch.org/t/errors-when-using-num-workers-0-in-dataloader/97564/4
    https://github.com/pytorch/pytorch/issues/70344

    Returns
    -------
    int
        The number of workers.
    """
    # Windows or MacOS: no multiprocessing.
    if platform.system() in ["Windows", "Darwin"]:
        return 0
    # Linux: scale the number of workers by the number of GPUs (if present).
    try:
        n_cpu = len(psutil.Process().cpu_affinity())
    except AttributeError:
        n_cpu = os.cpu_count()
    return (
        n_cpu // n_gpu if (n_gpu := torch.cuda.device_count()) > 1 else n_cpu
    )


def split_version(version: str) -> Tuple[str, str, str]:
    """
    Split the version into its semantic versioning components.

    Parameters
    ----------
    version : str
        The version number.

    Returns
    -------
    major : str
        The major release.
    minor : str
        The minor release.
    patch : str
        The patch release.
    """
    version_regex = re.compile(r"(\d+)\.(\d+)\.*(\d*)(?:.dev\d+.+)?")
    return tuple(g for g in version_regex.match(version).groups())



"""pytorch-lightning LightningDataModules for various tasks."""
import os

import torch
import numpy as np
from pytorch_lightning import LightningDataModule

from .hdf5 import SpectrumIndex, AnnotatedSpectrumIndex
from .datasets import SpectrumDataset, AnnotatedSpectrumDataset


class SpectrumDataModule(LightningDataModule):
    """Data loaders for mass spectrometry data.

    Parameters
    ----------
    train_index : SpectrumIndex or AnnotatedSpectrumIndex
        The spectrum index file for training.
    val_index : SpectrumIndex or AnnotatedSpectrumIndex
        The spectrum index file for validation.
    test_index : SpectrumIndex or AnnotatedSpectrumIndex
        The spectrum index file for testing.
    batch_size : int, optional
        The batch size to use for training and evaluations
    n_peaks : int, optional
        Keep only the top-n most intense peaks in any spectrum. ``None``
        retains all of the peaks.
    min_mz : float, optional
        The minimum m/z to include. The default is 140 m/z, in order to
        exclude TMT and iTRAQ reporter ions.
    num_workers : int, optional
        The number of workers to use for data loading. By default, the number
        of available CPU cores on the current machine is used.
    shuffle : Boolean
        Whether the batches are shuffled or not for training. The batches are
        not shuffled for validation or testing.
    """

    def __init__(
        self,
        train_index=None,
        val_index=None,
        test_index=None,
        batch_size=128,
        n_peaks=200,
        min_mz=140,
        preprocessing_fn=None,
        num_workers=None,
        shuffle=True,
    ):
        """Initialize the PairedSpectrumDataModule."""
        super().__init__()
        self.train_index = train_index
        self.val_index = val_index
        self.test_index = test_index
        self.batch_size = batch_size
        self.n_peaks = n_peaks
        self.min_mz = min_mz
        self.preprocessing_fn = preprocessing_fn
        self.num_workers = num_workers
        self.train_dataset = None
        self.valid_dataset = None
        self.test_dataset = None
        self.shuffle = shuffle

        if self.num_workers is None:
            self.num_workers = os.cpu_count()

    def setup(self, stage=None):
        """Set up the PyTorch Datasets.

        Parameters
        ----------
        stage : str {"fit", "validate", "test"}
            The stage indicating which Datasets to prepare. All are prepared
            by default.
        """
        if stage in (None, "fit", "validate"):
            if self.train_index is not None:
                self.train_dataset = self._make_dataset(self.train_index)

            if self.val_index is not None:
                self.val_dataset = self._make_dataset(self.val_index)

        if stage in (None, "test"):
            if self.test_index is not None:
                self.test_dataset = self._make_dataset(self.test_index)

    def _make_dataset(self, index, random_state=None):
        """Create a PyTorch Dataset.

        Parameters
        ----------
        index : SpectrumIndex or AnnotatedSpectrumIndex

        Returns
        -------
        SpectrumDataset or AnnotatedSpectrumDataset
            The instantiated dataset.
        """
        if isinstance(index, AnnotatedSpectrumIndex):
            dataset_class = AnnotatedSpectrumDataset
        elif isinstance(index, SpectrumIndex):
            dataset_class = SpectrumDataset
        else:
            raise ValueError(
                "index must be a SpectrumIndex or AnnotatedSpectrumIndex."
            )

        return dataset_class(
            index,
            n_peaks=self.n_peaks,
            min_mz=self.min_mz,
            preprocessing_fn=self.preprocessing_fn,
        )

    def _make_loader(self, dataset, shuffle):
        """Create a PyTorch DataLoader.

        Parameters
        ----------
        dataset : torch.utils.data.Dataset
            A PyTorch Dataset.
        shuffle : bool
            Shuffle the batches?

        Returns
        -------
        torch.utils.data.DataLoader
            A PyTorch DataLoader.
        """
        return torch.utils.data.DataLoader(
            dataset,
            batch_size=self.batch_size,
            collate_fn=dataset.collate_fn,
            pin_memory=True,
            num_workers=self.num_workers,
            shuffle=shuffle,
        )

    def train_dataloader(self):
        """Get the training DataLoader."""
        return self._make_loader(self.train_dataset, self.shuffle)

    def val_dataloader(self):
        """Get the validation DataLoader."""
        return self._make_loader(self.val_dataset, False)

    def test_dataloader(self):
        """Get the test DataLoader."""
        return self._make_loader(self.test_dataset, False)



"""Training and testing functionality for the de novo peptide sequencing
model."""
import glob
import logging
import operator
import os
import tempfile
import uuid
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import pytorch_lightning as pl
import torch
from depthcharge.data import AnnotatedSpectrumIndex, SpectrumIndex
from pytorch_lightning.strategies import DDPStrategy

from .. import utils
from ..data import ms_io
from ..denovo.dataloaders import DeNovoDataModule
from ..denovo.model import Spec2Pep


logger = logging.getLogger("diffnovo-dia")


def predict(
    peak_path: str,
    peak_feature: str, 
    model_filename: str,
    config: Dict[str, Any],
    out_writer: ms_io.MztabWriter,
) -> None:
    """
    Predict peptide sequences with a trained DiffNovo_DIA model.

    Parameters
    ----------
   peak_path : str
        The path with peak files for predicting peptide sequences.
    peak_feature: str
        The path of dia feature file
    model_filename : str
        The file name of the model weights (.ckpt file).
    config : Dict[str, Any]
        The configuration options.
    out_writer : ms_io.MztabWriter
        The mzTab writer to export the prediction results.
    """
    _execute_existing(peak_path, peak_feature, model_filename, config, False, out_writer)


def evaluate(
        peak_path: str, peak_feature: str, model_filename: str, config: Dict[str, Any], out_writer
) -> None:
    """
    Evaluate peptide sequence predictions from a trained DiffNovo-DIA model.

    Parameters
    ----------
    peak_path : str
        The path with peak files for predicting peptide sequences.
    peak_feature: str
        The path of dia feature file
    model_filename : str
        The file name of the model weights (.ckpt file).
    config : Dict[str, Any]
        The configuration options.
    """
    _execute_existing(peak_path, peak_feature, model_filename, config, True, out_writer)


def _execute_existing(
    peak_path: str,
    peak_feature: str, 
    model_filename: str,
    config: Dict[str, Any],
    annotated: bool,
    out_writer: Optional[ms_io.MztabWriter] = None,
) -> None:
    """
    Predict peptide sequences with a trained DiffNovo_DIA model with/without
    evaluation.

    Parameters
    ----------
    peak_path : str
        The path with peak files for predicting peptide sequences.
    model_filename : str
        The file name of the model weights (.ckpt file).
    config : Dict[str, Any]
        The configuration options.
    annotated : bool
        Whether the input peak files are annotated (execute in evaluation mode)
        or not (execute in prediction mode only).
    out_writer : Optional[ms_io.MztabWriter]
        The mzTab writer to export the prediction results.
    """
    # Load the trained model.
    if not os.path.isfile(model_filename):
        logger.error(
            "Could not find the trained model weights at file %s",
            model_filename,
        )
        raise FileNotFoundError("Could not find the trained model weights")
    model = Spec2Pep().load_from_checkpoint(
        model_filename,
        dim_model=config["dim_model"],
        n_head=config["n_head"],
        dim_feedforward=config["dim_feedforward"],
        n_layers=config["n_layers"],
        dropout=config["dropout"],
        dim_intensity=config["dim_intensity"],
        custom_encoder=config["custom_encoder"],
        max_length=config["max_length"],
        residues=config["residues"],
        max_charge=config["max_charge"],
        precursor_mass_tol=config["precursor_mass_tol"],
        isotope_error_range=config["isotope_error_range"],
        n_beams=config["n_beams"],
        n_log=config["n_log"],
        out_writer=out_writer,
    )
    # Read the MS/MS spectra for which to predict peptide sequences.
    if annotated:
        peak_ext = (".mgf", ".h5", ".hdf5")
    else:
        peak_ext = (".mgf", ".mzml", ".mzxml", ".h5", ".hdf5")
    if len(peak_filenames := _get_peak_filenames(peak_path, peak_ext)) == 0:
        logger.error("Could not find peak files from %s", peak_path)
        raise FileNotFoundError("Could not find peak files")
    peak_is_index = any(
        [os.path.splitext(fn)[1] in (".h5", ".hdf5") for fn in peak_filenames]
    )
    if peak_is_index and len(peak_filenames) > 1:
        logger.error("Multiple HDF5 spectrum indexes specified")
        raise ValueError("Multiple HDF5 spectrum indexes specified")
    tmp_dir = tempfile.TemporaryDirectory()
    if peak_is_index:
        idx_filename, peak_filenames = peak_filenames[0], None
    else:
        idx_filename = os.path.join(tmp_dir.name, f"{uuid.uuid4().hex}.hdf5")
    SpectrumIdx = AnnotatedSpectrumIndex if annotated else SpectrumIndex
    valid_charge = np.arange(1, config["max_charge"] + 1)
    index = SpectrumIdx(
        idx_filename, peak_filenames, peak_feature, valid_charge=valid_charge
    )
    # Initialize the data loader.
    loaders = DeNovoDataModule(
        test_index=index,
        n_peaks=config["n_peaks"],
        min_mz=config["min_mz"],
        max_mz=config["max_mz"],
        min_intensity=config["min_intensity"],
        remove_precursor_tol=config["remove_precursor_tol"],
        n_workers=config["n_workers"],
        batch_size=config["predict_batch_size"],
    )
    loaders.setup(stage="test", annotated=annotated)

    # Create the Trainer object.
    trainer = pl.Trainer(
        accelerator="auto",
        auto_select_gpus=True,
        devices=_get_devices(),
        logger=config["logger"],
        max_epochs=config["max_epochs"],
        num_sanity_val_steps=config["num_sanity_val_steps"],
        strategy=_get_strategy(),
    )
    # Run the model with/without validation.
    run_trainer = trainer.validate if annotated else trainer.predict
    run_trainer(model, loaders.test_dataloader())
    # Clean up temporary files.
    tmp_dir.cleanup()


def train(
    peak_path: str,
    peak_feature: str,
    peak_path_val: str,
    peak_feature_val: str, 
    model_filename: str,
    config: Dict[str, Any],
) -> None:
    """
    Train a DiffNovo-DIA model.

    The model can be trained from scratch or by continuing training an existing
    model.

    Parameters
    ----------
    peak_path : str
        The path with peak files to be used as training data.
    peak_path_val : str
        The path with peak files to be used as validation data.
    model_filename : str
        The file name of the model weights (.ckpt file).
    config : Dict[str, Any]
        The configuration options.
    """
    # Read the MS/MS spectra to use for training and validation.
    ext = (".mgf", ".h5", ".hdf5")
    if len(train_filenames := _get_peak_filenames(peak_path, ext)) == 0:
        logger.error("Could not find training peak files from %s", peak_path)
        raise FileNotFoundError("Could not find training peak files")
    train_is_index = any(
        [os.path.splitext(fn)[1] in (".h5", ".hdf5") for fn in train_filenames]
    )
    if train_is_index and len(train_filenames) > 1:
        logger.error("Multiple training HDF5 spectrum indexes specified")
        raise ValueError("Multiple training HDF5 spectrum indexes specified")
    if (
        peak_path_val is None
        or len(val_filenames := _get_peak_filenames(peak_path_val, ext)) == 0
    ):
        logger.error(
            "Could not find validation peak files from %s", peak_path_val
        )
        raise FileNotFoundError("Could not find validation peak files")
    val_is_index = any(
        [os.path.splitext(fn)[1] in (".h5", ".hdf5") for fn in val_filenames]
    )
    if val_is_index and len(val_filenames) > 1:
        logger.error("Multiple validation HDF5 spectrum indexes specified")
        raise ValueError("Multiple validation HDF5 spectrum indexes specified")
    tmp_dir = tempfile.TemporaryDirectory()
    if train_is_index:
        train_idx_fn, train_filenames = train_filenames[0], None
    else:
        train_idx_fn = os.path.join(tmp_dir.name, f"{uuid.uuid4().hex}.hdf5")
    valid_charge = np.arange(1, config["max_charge"] + 1)
    train_index = AnnotatedSpectrumIndex(
        train_idx_fn, train_filenames, peak_feature, valid_charge=valid_charge
    )
    if val_is_index:
        val_idx_fn, val_filenames = val_filenames[0], None
    else:
        val_idx_fn = os.path.join(tmp_dir.name, f"{uuid.uuid4().hex}.hdf5")
    val_index = AnnotatedSpectrumIndex(
        val_idx_fn, val_filenames, peak_feature_val, valid_charge=valid_charge
    )
    # Initialize the data loaders.
    dataloader_params = dict(
        batch_size=config["train_batch_size"],
        n_peaks=config["n_peaks"],
        min_mz=config["min_mz"],
        max_mz=config["max_mz"],
        min_intensity=config["min_intensity"],
        remove_precursor_tol=config["remove_precursor_tol"],
        n_workers=config["n_workers"],
    )
    train_loader = DeNovoDataModule(
        train_index=train_index, **dataloader_params
    )
    train_loader.setup()
    val_loader = DeNovoDataModule(valid_index=val_index, **dataloader_params)
    val_loader.setup()
    # Initialize the model.
    model_params = dict(
        dim_model=config["dim_model"],
        vocab_size=config["vocab_size"],
        win_size=config["win_size"],
        num_ions=config["num_ions"],
        num_neighbor=config["num_neighbors"],
        n_head=config["n_head"],
        dim_feedforward=config["dim_feedforward"],
        n_layers=config["n_layers"],
        dropout=config["dropout"],
        dim_intensity=config["dim_intensity"],
        custom_encoder=config["custom_encoder"],
        max_length=config["max_length"],
        residues=config["residues"],
        max_charge=config["max_charge"],
        precursor_mass_tol=config["precursor_mass_tol"],
        isotope_error_range=config["isotope_error_range"],
        n_beams=config["n_beams"],
        n_log=config["n_log"],
        tb_summarywriter=config["tb_summarywriter"],
        warmup_iters=config["warmup_iters"],
        max_iters=config["max_iters"],
        lr=config["learning_rate"],
        weight_decay=config["weight_decay"],
    )
    if config["train_from_scratch"]:
        model = Spec2Pep(**model_params)
    else:
        if not os.path.isfile(model_filename):
            logger.error(
                "Could not find the model weights at file %s to continue "
                "training",
                model_filename,
            )
            raise FileNotFoundError(
                "Could not find the model weights to continue training"
            )
        model = Spec2Pep().load_from_checkpoint(model_filename, **model_params)
    # Create the Trainer object and (optionally) a checkpoint callback to
    # periodically save the model.
    if config["save_model"]:
        callbacks = [
            pl.callbacks.ModelCheckpoint(
                dirpath=config["model_save_folder_path"],
                save_top_k=-1,
                save_weights_only=config["save_weights_only"],
                every_n_train_steps=config["every_n_train_steps"],
            )
        ]
    else:
        callbacks = None

    trainer = pl.Trainer(
        accelerator="auto",
        auto_select_gpus=True,
        callbacks=callbacks,
        devices=_get_devices(),
        logger=config["logger"],
        max_epochs=config["max_epochs"],
        num_sanity_val_steps=config["num_sanity_val_steps"],
        strategy=_get_strategy(),
    )
    # Train the model.
    trainer.fit(
        model, train_loader.train_dataloader(), val_loader.val_dataloader()
    )
    # Clean up temporary files.
    tmp_dir.cleanup()


def _get_peak_filenames(
    path: str, supported_ext: Iterable[str] = (".mgf",)
) -> List[str]:
    """
    Get all matching peak file names from the path pattern.

    Performs cross-platform path expansion akin to the Unix shell (glob, expand
    user, expand vars).

    Parameters
    ----------
    path : str
        The path pattern.
    supported_ext : Iterable[str]
        Extensions of supported peak file formats. Default: MGF.

    Returns
    -------
    List[str]
        The peak file names matching the path pattern.
    """
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    return [
        fn
        for fn in glob.glob(path, recursive=True)
        if os.path.splitext(fn.lower())[1] in supported_ext
    ]


def _get_strategy() -> Optional[DDPStrategy]:
    """
    Get the strategy for the Trainer.

    The DDP strategy works best when multiple GPUs are used. It can work for
    CPU-only, but definitely fails using MPS (the Apple Silicon chip) due to
    Gloo.

    Returns
    -------
    Optional[DDPStrategy]
        The strategy parameter for the Trainer.
    """
    if torch.cuda.device_count() > 1:
        return DDPStrategy(find_unused_parameters=False, static_graph=True)

    return None


def _get_devices() -> Union[int, str]:
    """
    Get the number of GPUs/CPUs for the Trainer to use.

    Returns
    -------
    Union[int, str]
        The number of GPUs/CPUs to use, or "auto" to let PyTorch Lightning
        determine the appropriate number of devices.
    """
    if any(
        operator.attrgetter(device + ".is_available")(torch)()
        for device in ["cuda", "backends.mps"]
    ):
        return -1
    elif not (n_workers := utils.n_workers()):
        return "auto"
    else:
        return n_workers



"""Package version information."""
from typing import Optional


def _get_version() -> Optional[str]:
    """
    Retrieve the package version.

    Using setuptools-scm dynamically sets the package version depending on the
    latest git release and commits since that point. This function returns the
    current version number.

    Returns
    -------
    Optional[str]
        The package version number. If not version is found, returns `None`.
    """
    try:
        # Fast, but only works in Python 3.8+.
        from importlib.metadata import version, PackageNotFoundError

        try:
            return version("diffnovo_dia")
        except PackageNotFoundError:
            return None
    except ImportError:
        # Slow, but works for all Python 3+.
        from pkg_resources import get_distribution, DistributionNotFound

        try:
            return get_distribution("diffnovo-dia").version
        except DistributionNotFound:
            return None
